"""Models for surveys with the Scintrex CG-5 gravity meters.

This module contains all models that are specifically required to
handle observation with the Scintrex CG-5 gravity meter.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd
import numpy as np
import re
import datetime as dt

from gravtools.models.exceptions import InvaliFileContentError
from gravtools import settings


class DataCursor:
    """Data cursor for matplotlib plot. X and Y coordinates are printed life to the plot canvas.

    From: https://stackoverflow.com/questions/4652439/is-there-a-matplotlib-equivalent-of-matlabs-datacursormode
    """
    # text_template = 'x: %0.2f\ny: %0.2f'
    x, y = 0.0, 0.0
    xoffset, yoffset = -20, 20
    text_template = 'x: %0.2f\ny: %0.2f'

    def __init__(self, ax):
        self.ax = ax
        self.annotation = ax.annotate(self.text_template,
                                      xy=(self.x, self.y), xytext=(self.xoffset, self.yoffset),
                                      textcoords='offset points', ha='right', va='bottom',
                                      bbox=dict(boxstyle='round,pad=0.5', fc='yellow', alpha=0.5),
                                      arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0')
                                      )
        self.annotation.set_visible(False)

    def __call__(self, event):
        self.event = event
        # xdata, ydata = event.artist.get_data()
        # self.x, self.y = xdata[event.ind], ydata[event.ind]
        self.x, self.y = event.mouseevent.xdata, event.mouseevent.ydata
        if self.x is not None:
            self.annotation.xy = self.x, self.y
            self.annotation.set_text(self.text_template % (self.x, self.y))
            self.annotation.set_visible(True)
            event.canvas.draw()


class CG5SurveyParameters:
    """CG-5 Survey parameters.

    Scintrex CG-5 survey parameters from the 'CG-5 SURVEY' block in
    the observation file (text format).
    The class is initialized either by the :py:meth:`.__init__`,
    method by passing the attributes as keyword arguments directly,
    or by the class method :py:meth:`.populate_from_obs_file_string`
    that parses the content of a CG-5 observation file (txt).

    Attributes
    ----------
    survey_name : str
        Name of gravity survey.
    client : str
        Client of survey.
    operator : str
        Name of the survey's operator.
    long_deg : float
        Geographical longitude at begin of survey [°].
    lat_deg : float
        Geographical latitude at begin of survey [°].
    zone : str
        Timezone of all time records.
    date_time : datetime object
        Start epoch of the survey.
     """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        *args
            Variable length argument list.
        **kwargs : dict
            Keyword arguments that are parsed to class attributes.
        """
        self.survey_name = kwargs.get('survey_name', '')  # string
        self.instrument_sn = kwargs.get('instrument_sn', '')  # string
        self.client = kwargs.get('client', '')  # string
        self.operator = kwargs.get('operator', '')  # string
        self.long_deg = kwargs.get('long_deg', np.nan)  # float
        self.lat_deg = kwargs.get('lat_deg', np.nan)  # float
        self.zone = kwargs.get('zone', '')  # string
        self.date_time = kwargs.get('date_time', None)  # datetime object (timezone aware)

    @classmethod
    def create_from_obs_file_string(cls, str_obs_file):
        """ Create instance of class :py:class:`.CG5SurveyParameters` by
        parsing a CG-5 observation file string.

        Parameters
        ----------
        str_obs_file : str
            Content of the CG-5 observation file (.txt).

        Returns
        -------
        Object: :py:class:`.CG5SurveyParameters`
            Initialized class instance.
        """

        # Parse observation file string:
        # expr = r'(\/\tCG-5 SURVEY\s*\n\/\s+Survey name:\s*(?P<survey_name>\S+)\s*\n\/\tInstrument S\/N:\s*(?P<instrument_sn>\S+)\s*\n\/\tClient:\s*(?P<client>\S+)\s*\n\/\tOperator:\s*(?P<operator>\S+)\s*\n\/\tDate:\s*(?P<date_year>\d{4})\/\s*(?P<date_month>\d{1,2})\/\s*(?P<date_day>\d{1,2})\s*\n\/\tTime:\s*(?P<time_hour>\d{2}):(?P<time_minu>\d{2}):(?P<time_sec>\d{2})\s*\n\/\tLONG:\s*(?P<long_num>\d*\.?\d*)\s+(?P<long_dir>[E|N|S|W])\s*\n\/\tLAT:\s*(?P<lat_num>\d*\.?\d*)\s+(?P<lat_dir>[E|N|S|W])\s*\n\/\tZONE:\s*(?P<zone>\d*)\s*\n\/\tGMT DIFF.:\s*(?P<gmt_diff>\d*\.?\d*))+\s*\n'
        expr = r'(\/\tCG-5 SURVEY\s*\n\/\s+Survey name:\s*(?P<survey_name>\S+)\s*\n' \
               r'\/\tInstrument S\/N:\s*(?P<instrument_sn>\S+)\s*\n' \
               r'\/\tClient:\s*(?P<client>\S+)\s*\n' \
               r'\/\tOperator:\s*(?P<operator>\S+)\s*\n' \
               r'\/\tDate:\s*(?P<date_year>\d{4})\/\s*(?P<date_month>\d{1,2})\/\s*(?P<date_day>\d{1,2})\s*\n' \
               r'\/\tTime:\s*(?P<time_hour>\d{2}):(?P<time_minu>\d{2}):(?P<time_sec>\d{2})\s*\n' \
               r'\/\tLONG:\s*(?P<long_num>\d*\.?\d*)\s+(?P<long_dir>[E|N|S|W])\s*\n' \
               r'\/\tLAT:\s*(?P<lat_num>\d*\.?\d*)\s+(?P<lat_dir>[E|N|S|W])\s*\n' \
               r'\/\tZONE:\s*(?P<zone>\d*)\s*\n' \
               r'\/\tGMT DIFF.:\s*(?P<gmt_diff>\d*\.?\d*))+\s*\n'

        # Read survey blocks from obs file string (only one block allowed!):
        survey_count = 0  # number of survey blocks in obs file string
        for survey_block in re.finditer(expr, str_obs_file):
            survey_dict = survey_block.groupdict()
            survey_count += 1

        if survey_count == 1:  # OK => Parse data in string:
            # Handle geographic locations:
            longitude_deg = float(survey_dict['long_num'])
            if survey_dict['long_dir'] == "W":
                longitude_deg = -longitude_deg
            latitude_deg = float(survey_dict['lat_num'])
            if survey_dict['lat_dir'] == "S":
                latitude_deg = -latitude_deg
            instrument_sn = survey_dict['instrument_sn']  # instrument serial number

            return cls(survey_name=survey_dict['survey_name'],
                       instrument_sn=instrument_sn,
                       client=survey_dict['client'],
                       operator=survey_dict['operator'],
                       long_deg=longitude_deg,
                       lat_deg=latitude_deg,
                       zone=survey_dict['zone'],
                       date_time=dt.datetime(int(survey_dict['date_year']),
                                             int(survey_dict['date_month']),
                                             int(survey_dict['date_day']),
                                             int(survey_dict['time_hour']),
                                             int(survey_dict['time_sec']),
                                             tzinfo=dt.timezone(dt.timedelta(hours=float(survey_dict['gmt_diff'])))
                                             )
                       )
        elif survey_count == 0:  # Not available
            return cls()  # Initialize with default values
        else:  # More than 1 block found => Error!
            raise InvaliFileContentError('{} "CG-5 SURVEY" blocks found in observation file, '
                                         'but only one expected.'.format(survey_count))

        # Error Msg, wenn der Block mehr als einmal gefunden wird.


class CG5SetupParameters:
    """CG-5 Survey parameters.

    Scintrex CG-5 Setup parameters from the 'CG-5 SETUP PARAMETERS'
    block in the observation file (text format).
    The class is initialized either by the :py:meth:`.__init__`,
    method by passing the attributes as keyword arguments directly,
    or by the class method :py:meth:`.populate_from_obs_file_string`
    that parses the content of a CG-5 observation file (txt).

    Attributes
    ----------
    gcal1 : float
        Calibration factor GCAL1.
    tiltxs : float
        XXXXXXXX
    tiltys : float
       XXXXXXXX
    tiltxo : float
        XXXXXXXX
    tiltyo : float
        XXXXXXXX
    tempco : float
        XXXXXXXX
    drift : float
        Linear drift factor (long-term drift).
    drift_date_time_start : datetime object (TZ aware)
        Start epoch for the determination of th linear long-term drift.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        *args
            Variable length argument list.
        **kwargs : dict
            Keyword arguments that are parsed to class attributes.
        """
        self.gcal1 = kwargs.get('gcal1', np.nan)  # float
        self.tiltxs = kwargs.get('tiltxs', np.nan)  # float
        self.tiltys = kwargs.get('tiltys', np.nan)  # string
        self.tiltxo = kwargs.get('tiltxo', np.nan)  # float
        self.tiltyo = kwargs.get('tiltyo', np.nan)  # float
        self.tempco = kwargs.get('tempco', np.nan)  # float
        self.drift = kwargs.get('drift', np.nan)  # float
        self.drift_date_time_start = kwargs.get('drift_date_time_start', None)  # datetime object (timezone aware)

    @classmethod
    def create_from_obs_file_string(cls, str_obs_file):
        """ Create instance of class :py:class:`.CG5SetupParameters` by
        parsing a CG-5 observation file string.

        Parameters
        ----------
        str_obs_file : str
            Content of the CG-5 observation file (.txt).

        Returns
        -------
        Object: :py:class:`.CG5SetupParameters`
            Initializes class instance.
        """

        # Parse observation file string:
        expr = r"\/\tCG-5 SETUP PARAMETERS\s*\n\/\s+Gref:\s*(?P<gref>\S+)\s*\n\/\s+Gcal1:\s*(" \
               r"?P<gcal1>\S+)\s*\n\/\s+TiltxS:\s*(?P<tiltxs>\S+)\s*\n\/\s+TiltyS:\s*(" \
               r"?P<tiltys>\S+)\s*\n\/\s+TiltxO:\s*(?P<tiltxo>\S+)\s*\n\/\s+TiltyO:\s*(" \
               r"?P<tiltyo>\S+)\s*\n\/\s+Tempco:\s*(?P<tempco>\S+)\s*\n\/\s+Drift:\s*(" \
               r"?P<drift>\S+)\s*\n\/\s+DriftTime Start:\s*(?P<drift_time_start>\S+)\s*\n\/\s+DriftDate Start:\s*(" \
               r"?P<drift_date_start>\S+)\s*\n"

        # Read setup parameters blocks from obs file string (only one block allowed!):
        block_count = 0  # number of survey blocks in obs file string
        for survey_block in re.finditer(expr, str_obs_file):
            setup_dict = survey_block.groupdict()
            block_count += 1

        if block_count == 1:  # OK => Parse data in string:

            return cls(gref=float(setup_dict['gref']),
                       gcal1=float(setup_dict['gcal1']),
                       tiltxs=float(setup_dict['tiltxs']),
                       tiltys=float(setup_dict['tiltys']),
                       tiltxo=float(setup_dict['tiltxo']),
                       tiltyo=float(setup_dict['tiltyo']),
                       tempco=float(setup_dict['tempco']),
                       drift=float(setup_dict['drift']),
                       drift_date_time_start=dt.datetime.strptime(
                           setup_dict["drift_date_start"] + setup_dict["drift_time_start"],
                           "%Y/%m/%d%H:%M:%S")
                       )
        elif block_count == 0:  # Not available
            return cls()  # Initialize with default values
        else:  # More than 1 block found => Error!
            raise InvaliFileContentError('{} "CG-5 SETUP PARAMETERS" in observation file found, but maximum one '
                                         'expected.'.format(block_count))

    # def is_valid(self) -> bool:
    #     """???"""
    #     # TODO
    #     return True


class CG5OptionsParameters:
    """CG-5 instrumental options (filters, corrections, output).

    Scintrex CG-5 options from the 'CG-5 OPTIONS' block in
    the observation file (text format).
    The class is initialized either by the :py:meth:`.__init__`,
    method by passing the attributes as keyword arguments directly,
    or by the class method :py:meth:`.populate_from_obs_file_string`
    that parses the content of a CG-5 observation file (txt).

    Attributes
    ----------
    tide_correction : bool
        Tide correction (on/off).
    cont_tilt : bool
        Continuous tilt correction (on/off).
    auto_rejection : bool
       Auto rejection of outliers (on/off).
    terrain_correction : bool
        Terrain correction (on/off).
    seismic_filter : bool
        Seismic filter (on/off).
    raw_data : bool
        Raw data output (on/off).
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        *args
            Variable length argument list.
        **kwargs : dict
            Keyword arguments that are parsed to class attributes.
        """
        self.tide_correction = kwargs.get('tide_correction', None)  # bool
        self.cont_tilt = kwargs.get('cont_tilt', None)  # bool
        self.auto_rejection = kwargs.get('auto_rejection', None)  # bool
        self.terrain_correction = kwargs.get('terrain_correction', None)  # bool
        self.seismic_filter = kwargs.get('seismic_filter', None)  # bool
        self.raw_data = kwargs.get('raw_data', None)  # bool
        #
        # self.tide_correction = kwargs.get('tide_correction', False)  # bool
        # self.cont_tilt = kwargs.get('cont_tilt', False)  # bool
        # self.auto_rejection = kwargs.get('auto_rejection', False)  # bool
        # self.terrain_correction = kwargs.get('terrain_correction', False)  # bool
        # self.seismic_filter = kwargs.get('seismic_filter', False)  # bool
        # self.raw_data = kwargs.get('raw_data', False)  # bool

    @classmethod
    def create_from_obs_file_string(cls, str_obs_file):
        """ Create instance of class :py:class:`.CG5OptionsParameters` by
        parsing a CG-5 observation file string.

        Parameters
        ----------
        str_obs_file : str
            Content of the CG-5 observation file (.txt).

        Returns
        -------
        Object: :py:class:`.CG5OptionsParameters`
            Initializes class instance.
        """

        # Parse observation file string:
        expr = r"\/\tCG-5 OPTIONS\s*\n\/\s+Tide Correction:\s*(?P<tide_correction>\S+)\s*\n\/\s+Cont. Tilt:\s*(" \
               r"?P<cont_tilt>\S+)\s*\n\/\s+Auto Rejection:\s*(?P<auto_rejection>\S+)\s*\n\/\s+Terrain Corr.:\s*(" \
               r"?P<terrain_correction>\S+)\s*\n\/\s+Seismic Filter:\s*(?P<seismic_filter>\S+)\s*\n\/\s+Raw Data:\s*(" \
               r"?P<raw_data>\S+)\s*\n"

        # Read setup parameters blocks from obs file string (only one block allowed!):
        block_count = 0  # number of survey blocks in obs file string
        for options_block in re.finditer(expr, str_obs_file):
            options_dict = options_block.groupdict()
            block_count += 1

        if block_count == 1:  # OK => Parse data in string:

            return cls(tide_correction=options_dict['tide_correction'] == "YES",  # bool
                       cont_tilt=options_dict['cont_tilt'] == "YES",  # bool
                       auto_rejection=options_dict['auto_rejection'] == "YES",  # bool
                       terrain_correction=options_dict['terrain_correction'] == "YES",  # bool
                       seismic_filter=options_dict['seismic_filter'] == "YES",  # bool
                       raw_data=options_dict['raw_data'] == "YES",
                       )  # bool

        elif block_count == 0:  # Not available
            return cls()  # Initialize with default values
        else:  # More than 1 block found => Error!
            raise InvaliFileContentError('{} "CG-5 OPTIONS" in observation file found, but maximum one '
                                         'expected.'.format(block_count))

    # def is_valid(self) -> bool:
    #     """???"""
    #     # TODO
    #     return True


class CG5Survey:
    """CG-5 survey data.

    Class instances may contain the information available in the
    following sections of Scintrex CG-5 observation files (txt format):

    - Survey Parameter block (as instance of :py:obj:`.CG5SurveyParameters`)
    - Setup block (as instance of :py:obj:`.CG5SetupParameters`)
    - Options block (as instance of :py:obj:`.CG5OptionsParameters`)
    - Observations (as pandas dataframe)

    If the class is initialized without setting the observations file
    attribute (`obs_filename`), no observation file is load and the
    object is initialized empty.

    Attributes
    ----------
    obs_filename : str
        Name (and path) to CG-5 observation file (txt format)
    survey_parameters : :py:class:`.CG5SurveyParameters`
        Survey Parameter of the Parameters block in the observation file.
    setup_parameters :  :py:class:`.CG5SetupParameters`
        Setup Parameter of the Setup block in the observation file.
    options : :py:class:`.CG5OptionsParameters`
        Instrumental options from the Options block in the observation file.
    obs_df : pandas data frame
       Contains the actual observation data records with the colums defined in `self._OBS_DF_COLUMN_NAMES`.
    """

    # Column names of the dataframe containing tha actual observation data:
    _OBS_DF_COLUMN_NAMES = ('lat_deg',  # Latitude [deg] :
                            'lon_deg',  # Longitude [deg]
                            'alt_m',  # Altitude [m]
                            'g_mgal',  # Determined gravity (corrected) [mGal]
                            'sd_mgal',  # Standard deviation of determined gravity [mGal]
                            'tiltx',
                            'tilty',
                            'temp',
                            'tide',  # Tidal correction determined by the CG-5 [mGal]
                            'duration_sec',  # Duration of the current setup [sec]
                            'rej',  # Number of rejected single measurements
                            'time_str',  # Reference time = mid of setup with duration `duration_sec`) (dropped later)
                            'dec_time_date',
                            'terrain',  # Terrain correction [??]
                            'date',  # Date (dropped later)
                            'station_name',  # Station name : str
                            'dhf_m',  # Distance between instrument top and physical reference point [m]
                            'dhb_m',  # Distance between instrument top and ground [m]
                            'atm_pres_hpa',  # Measured atmospheric pressure [hPa]
                            'setup_id',  # Unique ID of this observation (=setup)
                            )
                            # obs_epoch : datetime object (added to df later)

    # Rename columns: df.rename(columns = {'$b':'B'}, inplace = True)

    # Non-numeric columns in the observation dataframe:
    _OBS_DF_NON_NUMERIC_COLUMNS = ['station_name', 'date', 'time_str']

    def __init__(self,
                 obs_filename='',
                 survey_parameters=CG5SurveyParameters(),
                 setup_parameters=CG5SetupParameters(),
                 options=CG5OptionsParameters()
                 ):
        """
        Parameters
        ----------
        obs_filename : str, optional
            Name (and path) to CG-5 observation file (txt format).
        survey_parameters : :py:class:`.CG5SurveyParameters`, optional
            Survey Parameter of the Parameters block in the observation file.
        setup_parameters :  :py:class:`.CG5SetupParameters`, optional
            Setup Parameter of the Setup block in the observation file.
        options : :py:class:`.CG5OptionsParameters`, optional
            Instrumental options from the Options block in the observation file.
        """
        self.obs_filename = obs_filename
        assert isinstance(survey_parameters, CG5SurveyParameters), \
            "survey_parameters is not an instance of CG5SurveyParameters"
        self.survey_parameters = survey_parameters
        assert isinstance(setup_parameters, CG5SetupParameters), \
            "setup_parameters is not an instance of CG5SetupParameters"
        self.setup_parameters = setup_parameters
        assert isinstance(options, CG5OptionsParameters), \
            "options is not an instance of CG5OptionsParameters"
        self.options = options

        # Read observation file, if a valid filename is available and valid. Otherwise, initialize obs_df as None.
        if self.obs_filename:
            self.read_obs_file(obs_filename)
        else:
            self.obs_df = None  # Initialize as None

    def __str__(self):
        if self.obs_df is None:
            return 'Empty CG-5 Survey.'
        else:
            if not self.survey_parameters.survey_name:
                return 'Unnamed CG-5 Survey with {} observations (file: {}).'.format(len(self.obs_df),
                                                                                     self.obs_filename.split('/')[-1])
            else:
                return 'CG-5 Survey "{}" with {} observations (file: {}).'.format(self.survey_parameters.survey_name,
                                                                                  len(self.obs_df),
                                                                                  self.obs_filename.split('/')[-1])

    @staticmethod
    def resolve_station_name(station_name_in):
        """Convert station name from Scintrex observation file
        (as Note) to the naming convention used in the output
        file (BEV conventions).

        Parameters
        ----------
        station_name_in : str
            Station name string as written to the observation file
            as note.

        Returns
        -------
        Corrected station name : str
        """
        station_name_in = station_name_in.upper()
        # Check first letter of name in order detect the station type:
        if station_name_in[0] == 'S':
            station_name_out = station_name_in
        elif station_name_in[0] == 'P':
            station_name_out = 'P  ' + station_name_in[1:]
        elif station_name_in[0] == 'T' and '.' in station_name_in:
            [str1_tmp, str2_tmp] = station_name_in.split('.')
            station_name_out = 'T{0:>4} {1:>3}'.format(str1_tmp[1:], str2_tmp)
        elif station_name_in[0] == 'N':
            station_name_out = station_name_in
        else:
            station_name_out = station_name_in.replace('.', '-')
        return station_name_out.upper()

    @staticmethod
    def get_dhb_dhf(dh_str):
        """Convert dhb and dhf from notes in the CG-5 observation file
        to an actual number. In the notes '.' is used instead of '-'.

        Parameters
        ----------
        dh_str : str
            Height difference [m] with '.' instead of '-'.

        Returns
        -------
        Height difference [m] : float
        """
        if dh_str.startswith('.'):
            dh_str = '-' + dh_str[1:]
        return float(dh_str)

    def read_obs_file(self, obs_filename):
        """Read CG-5 observation file (txt) and populate the object.

        Notes
        -----
        Ignore comment lines that start with "#".

        Parameters
        ----------
        obs_filename : str
            Name (and path) to CG-5 observation file (txt format).
        """
        COMMENT_MARKER = '#'

        self.obs_filename = obs_filename
        # with open(self.obs_filename, 'r') as content_file:
        #     str_obs_file = content_file.read()

        # Read in file and ignore comment lines:
        file_handle = open(self.obs_filename, 'r')
        lines = []
        for line in file_handle:
            line_tmp = line.strip()
            if not line_tmp.startswith(COMMENT_MARKER):
                lines.append(line_tmp)
        file_handle.close()
        str_obs_file = '\n'.join(lines)

        # Remove all or add one end-of-line symbols from end of string:
        # - Last character of string has to be a \n so that regex works correctly!
        number_of_trailing_eol_symbols = 0
        str_idx = -1
        while str_obs_file[str_idx] == '\n':
            number_of_trailing_eol_symbols += 1
            str_idx -= 1
        if number_of_trailing_eol_symbols == 0:
            str_obs_file += '\n'
        elif number_of_trailing_eol_symbols > 1:
            str_obs_file = str_obs_file[0:str_idx+2]

        # ### Match blocks with regex ###
        # CG-5 SURVEY block:
        self.survey_parameters = CG5SurveyParameters.create_from_obs_file_string(str_obs_file)
        self.setup_parameters = CG5SetupParameters.create_from_obs_file_string(str_obs_file)
        self.options = CG5OptionsParameters.create_from_obs_file_string(str_obs_file)

        # Get Observations
        # ### 3 possibilities: ###
        # Initialize empty dataframe and append observation blocks at each station
        # Warning: Better performance, when preparing the data as list (appending) and then converting to df at once.
        #  See: https://stackoverflow.com/questions/13784192/creating-an-empty-pandas-dataframe-then-filling-it

        obs_list = []  # Collect all obs data in this list and then convert to pd dataframe.

        # 1.) Station name & dbh=dhf
        expr = '\/\tNote:   \t(?P<station_name>\S+)\s+(?P<dh_cm>-?[.0-9]+)\s*[\r?\n](?P<obs_data>(?:\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s*[\r?\n])+)'
        for obs_block in re.finditer(expr, str_obs_file):
            obs_dict = obs_block.groupdict()
            station_name = self.resolve_station_name(obs_dict['station_name'])
            dhf_m = float(obs_dict['dh_cm']) * 1e-2
            dhb_m = float(obs_dict['dh_cm']) * 1e-2
            atm_pres_hpa = None
            lines = obs_dict['obs_data'].splitlines()
            # Create unique ID (= UNIX timestamp of first observation) for each setup on a station:
            #  - To distinguish multiple setups (with multiple observations each) on multiple stations
            time_str = lines[0].split()[-1] + ' ' + lines[0].split()[11]
            setup_id = int(dt.datetime.timestamp(dt.datetime.strptime(time_str, "%Y/%m/%d %H:%M:%S")))

            for line in lines:
                line_items = line.split()
                line_items.append(station_name)
                line_items.append(dhf_m)
                line_items.append(dhb_m)
                line_items.append(atm_pres_hpa)
                line_items.append(setup_id)
                obs_list.append(line_items)

        # 2.) Station name & dbh=dhf & pressure
        expr = '\/\tNote:   \t(?P<station_name>\S+)\s+(?P<dh_cm>-?[.0-9]+)\s*[\r?\n](?P<obs_data>(?:\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s*[\r?\n])+)\/\tNote:   \t(?P<pres>[0-9]{3,4}[.]{0,1}[0-9]*)'
        for obs_block in re.finditer(expr, str_obs_file):
            obs_dict = obs_block.groupdict()
            station_name = self.resolve_station_name(obs_dict['station_name'])
            dhf_m = float(obs_dict['dh_cm']) * 1e-2
            dhb_m = float(obs_dict['dh_cm']) * 1e-2
            atm_pres_hpa = float(obs_dict['pres'])
            lines = obs_dict['obs_data'].splitlines()
            # Create unique ID (= UNIX timestamp of first observation) for each setup on a station:
            #  - To distinguish multiple setups (with multiple observations each) on multiple stations
            time_str = lines[0].split()[-1] + ' ' + lines[0].split()[11]
            setup_id = int(dt.datetime.timestamp(dt.datetime.strptime(time_str, "%Y/%m/%d %H:%M:%S")))

            for line in lines:
                line_items = line.split()
                line_items.append(station_name)
                line_items.append(dhf_m)
                line_items.append(dhb_m)
                line_items.append(atm_pres_hpa)
                line_items.append(setup_id)
                obs_list.append(line_items)

        # 3.) Station name & dhb & dhf
        expr = '\/\tNote:   \t(?P<station_name>\S+)\s+(?P<dhb_cm>-?[.0-9]+)\s+(?P<dhf_cm>-?[.0-9]+)\s*[\r?\n](?P<obs_data>(?:\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s*[\r?\n])+)'
        for obs_block in re.finditer(expr, str_obs_file):
            obs_dict = obs_block.groupdict()

            station_name = self.resolve_station_name(obs_dict['station_name'])
            dhf_m = float(obs_dict['dhf_cm']) * 1e-2
            dhb_m = float(obs_dict['dhb_cm']) * 1e-2
            atm_pres_hpa = None
            lines = obs_dict['obs_data'].splitlines()
            # Create unique ID (= UNIX timestamp of first observation) for each setup on a station:
            #  - To distinguish multiple setups (with multiple observations each) on multiple stations
            time_str = lines[0].split()[-1] + ' ' + lines[0].split()[11]
            setup_id = int(dt.datetime.timestamp(dt.datetime.strptime(time_str, "%Y/%m/%d %H:%M:%S")))

            for line in lines:
                line_items = line.split()
                line_items.append(station_name)
                line_items.append(dhf_m)
                line_items.append(dhb_m)
                line_items.append(atm_pres_hpa)
                line_items.append(setup_id)
                obs_list.append(line_items)

        # 4.) Station name & dhb & dhf & pressure
        # expr = "\/\s+Note:\s+(?P<station_name>\S+)\s+(?P<dhb_cm>\S+)\s+(?P<dhf_cm>\S+)\s*\n(?P<obs_data>(?:\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s*[\r\n])+)"
        expr = '\/\tNote:   \t(?P<station_name>\S+)\s+(?P<dhb_cm>-?[.0-9]+)\s+(?P<dhf_cm>-?[.0-9]+)\s*[\r?\n](?P<obs_data>(?:\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s*[\r?\n])+)\/\tNote:   \t(?P<pres>[0-9]{3,4}[.]{0,1}[0-9]*)'
        for obs_block in re.finditer(expr, str_obs_file):
            obs_dict = obs_block.groupdict()

            station_name = self.resolve_station_name(obs_dict['station_name'])
            dhf_m = float(obs_dict['dhf_cm']) * 1e-2
            dhb_m = float(obs_dict['dhb_cm']) * 1e-2
            atm_pres_hpa = float(obs_dict['pres'])
            lines = obs_dict['obs_data'].splitlines()
            # Create unique ID (= UNIX timestamp of first observation) for each setup on a station:
            #  - To distinguish multiple setups (with multiple observations each) on multiple stations
            time_str = lines[0].split()[-1] + ' ' + lines[0].split()[11]
            setup_id = int(dt.datetime.timestamp(dt.datetime.strptime(time_str, "%Y/%m/%d %H:%M:%S")))

            for line in lines:
                line_items = line.split()
                line_items.append(station_name)
                line_items.append(dhf_m)
                line_items.append(dhb_m)
                line_items.append(atm_pres_hpa)
                line_items.append(setup_id)
                obs_list.append(line_items)

        # Create pandas dataframe of prepared list:
        self.obs_df = pd.DataFrame(obs_list, columns=self._OBS_DF_COLUMN_NAMES)

        # Remove duplicates entries that were matched with and without pressure:
        setup_ids = self.obs_df['setup_id'].unique().tolist()
        setup_ids_diplicates = []
        for setup_id in setup_ids:
            tmp_filter = self.obs_df['setup_id'] == setup_id
            if (~self.obs_df.loc[tmp_filter, 'atm_pres_hpa'].isna()).any():  # Entries with pressure found
                setup_ids_diplicates.append(setup_id)

        if setup_ids_diplicates:
            tmp_filter = ~(self.obs_df['atm_pres_hpa'].isna() & self.obs_df['setup_id'].isin(setup_ids_diplicates))
            self.obs_df = self.obs_df.loc[tmp_filter].copy(deep=True)

        # Convert numeric columns to numeric dtypes:
        cols = self.obs_df.columns.drop(self._OBS_DF_NON_NUMERIC_COLUMNS)
        self.obs_df[cols] = self.obs_df[cols].apply(pd.to_numeric, errors='raise')
        # Sort observations by time and date and reset index:
        self.obs_df.sort_values(by='dec_time_date', inplace=True, ignore_index=True)
        # Convert date and time to datetime objects (aware, if UTC offset is available):
        if self.survey_parameters.date_time is not None:
            self.obs_df['obs_epoch'] = pd.to_datetime(
                self.obs_df['date'] + ' ' + self.obs_df['time_str'], format='%Y/%m/%d %H:%M:%S')
            self.obs_df['obs_epoch'] = self.obs_df['obs_epoch'].dt.tz_localize('UTC')  # Set timezone = UTC
            self.obs_df['obs_epoch'] = self.obs_df['obs_epoch'] + pd.Timedelta(self.survey_parameters.date_time.utcoffset())
        else:  # tz unaware time
            self.obs_df['obs_epoch'] = pd.to_datetime(
                self.obs_df['date'] + ' ' + self.obs_df['time_str'], format='%Y/%m/%d %H:%M:%S')

        self.obs_df.drop(columns=['time_str', 'date'], inplace=True)  # Drop columns that are not required any more

    def plot_g_values(self, station_names=None):
        """Plot g-values of selected or all stations in the df.

        Notes
        -----
        This method requires matplotlib as optional dependency!

        Parameters
        ----------
        station_names : list of str, optional
            List of names of stations for which the observations will be plotted.
            The default value is None which implied that the data of all stations
            is plotted.

        """
        # Check if obs data is available first:
        if self.obs_df is not None:
            # Get list of station names and loop over them:
            if station_names is None:
                station_names = self.obs_df['station_name'].unique()
            fig, ax = plt.subplots()
            for station_name in station_names:
                x = self.obs_df[self.obs_df['station_name'] == station_name].obs_epoch
                y = self.obs_df[self.obs_df['station_name'] == station_name].g_mgal * 1e3  # µGal
                ax.scatter(x, y, label=station_name)
            ax.legend()
            ax.grid(True)
            ax.set_xlabel('Time')
            ax.set_ylabel('g [µGal]')
            ax.set_title('survey: ' + self.survey_parameters.survey_name)

            # Data cursor:
            fig.canvas.mpl_connect('pick_event', DataCursor(plt.gca()))

            plt.show()


# Run as standalone program:
if __name__ == "__main__":
    import matplotlib.pyplot as plt
    path = settings.PATH_OBS_FILE_CG5 + settings.NAME_OBS_FILE_CG5
    s1 = CG5Survey()
    s1.read_obs_file(path)
    # s1.plot_g_values(['1-164-04', '1-164-12', '1-164-11'])
    # s1.plot_g_values(['TEST'])
    s1.plot_g_values()
else:
    pass



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gravtools/gui/dialog_autoselection_settings.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog_autoselection_settings(object):
    def setupUi(self, Dialog_autoselection_settings):
        Dialog_autoselection_settings.setObjectName("Dialog_autoselection_settings")
        Dialog_autoselection_settings.resize(310, 333)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(Dialog_autoselection_settings)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox_autoselect_thresholds = QtWidgets.QGroupBox(Dialog_autoselection_settings)
        self.groupBox_autoselect_thresholds.setObjectName("groupBox_autoselect_thresholds")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_autoselect_thresholds)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 2, 3, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 0, 3, 1, 1)
        self.spinBox_n = QtWidgets.QSpinBox(self.groupBox_autoselect_thresholds)
        self.spinBox_n.setProperty("value", 3)
        self.spinBox_n.setObjectName("spinBox_n")
        self.gridLayout.addWidget(self.spinBox_n, 3, 4, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 1, 5, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem2, 1, 3, 1, 1)
        self.spinBox_tilt = QtWidgets.QSpinBox(self.groupBox_autoselect_thresholds)
        self.spinBox_tilt.setProperty("value", 5)
        self.spinBox_tilt.setObjectName("spinBox_tilt")
        self.gridLayout.addWidget(self.spinBox_tilt, 1, 4, 1, 1)
        self.label = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 2, 1, 1)
        self.spinBox_sd = QtWidgets.QSpinBox(self.groupBox_autoselect_thresholds)
        self.spinBox_sd.setProperty("value", 10)
        self.spinBox_sd.setObjectName("spinBox_sd")
        self.gridLayout.addWidget(self.spinBox_sd, 0, 4, 1, 1)
        self.checkBox_tilt = QtWidgets.QCheckBox(self.groupBox_autoselect_thresholds)
        self.checkBox_tilt.setChecked(True)
        self.checkBox_tilt.setObjectName("checkBox_tilt")
        self.gridLayout.addWidget(self.checkBox_tilt, 1, 0, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_9.setAlignment(QtCore.Qt.AlignCenter)
        self.label_9.setWordWrap(False)
        self.label_9.setObjectName("label_9")
        self.gridLayout.addWidget(self.label_9, 3, 3, 1, 1)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem3, 2, 1, 1, 1)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem4, 0, 1, 1, 1)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem5, 1, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 0, 5, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 2, 5, 1, 1)
        self.checkBox_sd = QtWidgets.QCheckBox(self.groupBox_autoselect_thresholds)
        self.checkBox_sd.setChecked(True)
        self.checkBox_sd.setObjectName("checkBox_sd")
        self.gridLayout.addWidget(self.checkBox_sd, 0, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 2, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 2, 2, 1, 1)
        self.checkBox_delta_g = QtWidgets.QCheckBox(self.groupBox_autoselect_thresholds)
        self.checkBox_delta_g.setChecked(False)
        self.checkBox_delta_g.setObjectName("checkBox_delta_g")
        self.gridLayout.addWidget(self.checkBox_delta_g, 2, 0, 1, 1)
        self.spinBox_delta_g = QtWidgets.QSpinBox(self.groupBox_autoselect_thresholds)
        self.spinBox_delta_g.setProperty("value", 5)
        self.spinBox_delta_g.setObjectName("spinBox_delta_g")
        self.gridLayout.addWidget(self.spinBox_delta_g, 2, 4, 1, 1)
        self.checkBox_duration = QtWidgets.QCheckBox(self.groupBox_autoselect_thresholds)
        self.checkBox_duration.setChecked(True)
        self.checkBox_duration.setObjectName("checkBox_duration")
        self.gridLayout.addWidget(self.checkBox_duration, 4, 0, 1, 1)
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem6, 4, 1, 1, 1)
        spacerItem7 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem7, 4, 3, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 4, 2, 1, 1)
        self.label_8 = QtWidgets.QLabel(self.groupBox_autoselect_thresholds)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 4, 5, 1, 1)
        self.spinBox_duration = QtWidgets.QSpinBox(self.groupBox_autoselect_thresholds)
        self.spinBox_duration.setProperty("value", 80)
        self.spinBox_duration.setObjectName("spinBox_duration")
        self.gridLayout.addWidget(self.spinBox_duration, 4, 4, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.verticalLayout_3.addWidget(self.groupBox_autoselect_thresholds)
        self.groupBox_reference_data = QtWidgets.QGroupBox(Dialog_autoselection_settings)
        self.groupBox_reference_data.setObjectName("groupBox_reference_data")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_reference_data)
        self.verticalLayout.setObjectName("verticalLayout")
        self.radioButton_ref_data_reduced_observations = QtWidgets.QRadioButton(self.groupBox_reference_data)
        self.radioButton_ref_data_reduced_observations.setChecked(True)
        self.radioButton_ref_data_reduced_observations.setObjectName("radioButton_ref_data_reduced_observations")
        self.buttonGroup_autoselect_reference_data = QtWidgets.QButtonGroup(Dialog_autoselection_settings)
        self.buttonGroup_autoselect_reference_data.setObjectName("buttonGroup_autoselect_reference_data")
        self.buttonGroup_autoselect_reference_data.addButton(self.radioButton_ref_data_reduced_observations)
        self.verticalLayout.addWidget(self.radioButton_ref_data_reduced_observations)
        self.radioButton_ref_data_original_observations = QtWidgets.QRadioButton(self.groupBox_reference_data)
        self.radioButton_ref_data_original_observations.setChecked(False)
        self.radioButton_ref_data_original_observations.setObjectName("radioButton_ref_data_original_observations")
        self.buttonGroup_autoselect_reference_data.addButton(self.radioButton_ref_data_original_observations)
        self.verticalLayout.addWidget(self.radioButton_ref_data_original_observations)
        self.verticalLayout_3.addWidget(self.groupBox_reference_data)
        spacerItem8 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem8)

        self.retranslateUi(Dialog_autoselection_settings)
        QtCore.QMetaObject.connectSlotsByName(Dialog_autoselection_settings)

    def retranslateUi(self, Dialog_autoselection_settings):
        _translate = QtCore.QCoreApplication.translate
        Dialog_autoselection_settings.setWindowTitle(_translate("Dialog_autoselection_settings", "Autoselection Settings"))
        self.groupBox_autoselect_thresholds.setTitle(_translate("Dialog_autoselection_settings", "Autoselect options and thresholds"))
        self.spinBox_n.setToolTip(_translate("Dialog_autoselection_settings", "Stabilized g is derived as mean value of last n observations in a setup."))
        self.label_6.setToolTip(_translate("Dialog_autoselection_settings", "Instrument tilt"))
        self.label_6.setText(_translate("Dialog_autoselection_settings", "arcsec"))
        self.spinBox_tilt.setToolTip(_translate("Dialog_autoselection_settings", "Instrument tilt"))
        self.label.setToolTip(_translate("Dialog_autoselection_settings", "Standard deviation of g"))
        self.label.setText(_translate("Dialog_autoselection_settings", ">"))
        self.spinBox_sd.setToolTip(_translate("Dialog_autoselection_settings", "Standard deviation of g"))
        self.checkBox_tilt.setToolTip(_translate("Dialog_autoselection_settings", "Instrument tilt"))
        self.checkBox_tilt.setText(_translate("Dialog_autoselection_settings", "Tilt"))
        self.label_9.setToolTip(_translate("Dialog_autoselection_settings", "Stabilized g is derived as mean value of last n observations in a setup."))
        self.label_9.setText(_translate("Dialog_autoselection_settings", "n ="))
        self.label_4.setToolTip(_translate("Dialog_autoselection_settings", "Standard deviation of g"))
        self.label_4.setText(_translate("Dialog_autoselection_settings", "µGal"))
        self.label_5.setToolTip(_translate("Dialog_autoselection_settings", "Deviation of the stabilized g in a setup. "))
        self.label_5.setText(_translate("Dialog_autoselection_settings", "µGal"))
        self.checkBox_sd.setToolTip(_translate("Dialog_autoselection_settings", "Standard deviation of g"))
        self.checkBox_sd.setText(_translate("Dialog_autoselection_settings", "SD(g)"))
        self.label_2.setToolTip(_translate("Dialog_autoselection_settings", "Instrument tilt"))
        self.label_2.setText(_translate("Dialog_autoselection_settings", ">"))
        self.label_3.setToolTip(_translate("Dialog_autoselection_settings", "Deviation of the stabilized g in a setup. "))
        self.label_3.setText(_translate("Dialog_autoselection_settings", ">"))
        self.checkBox_delta_g.setToolTip(_translate("Dialog_autoselection_settings", "Deviation of the stabilized g in a setup. "))
        self.checkBox_delta_g.setText(_translate("Dialog_autoselection_settings", "Delta g"))
        self.spinBox_delta_g.setToolTip(_translate("Dialog_autoselection_settings", "Deviation of the stabilized g in a setup. "))
        self.checkBox_duration.setText(_translate("Dialog_autoselection_settings", "Duration"))
        self.label_7.setText(_translate("Dialog_autoselection_settings", "<"))
        self.label_8.setText(_translate("Dialog_autoselection_settings", "sec"))
        self.groupBox_reference_data.setTitle(_translate("Dialog_autoselection_settings", "Reference data"))
        self.radioButton_ref_data_reduced_observations.setText(_translate("Dialog_autoselection_settings", "Reduced observations"))
        self.radioButton_ref_data_original_observations.setText(_translate("Dialog_autoselection_settings", "Original observations"))



"""Model classes for PyQt's model view architecture.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt5.QtCore import QAbstractTableModel, Qt, QPersistentModelIndex
from PyQt5 import QtGui
from PyQt5.QtWidgets import QMessageBox, QStyledItemDelegate

from gravtools import settings

NONE_REPRESENTATION_IN_TABLE_VIEW = ''  # Representation of None values in table views in the GUI


class StationTableModel(QAbstractTableModel):
    """Model for displaying the station data in a table view (QTableView)."""

    # Columns that will be shown in the table view, if available in the data (Also defines the order of columns):
    # - keys: Actual names of the dataframe columns
    # - items: Header names for the Table View Widget
    _SHOW_COLUMNS_IN_TABLE_DICT = {
        'station_name': 'Station',
        'long_deg': 'Lon [°]',
        'lat_deg': 'Lat [°]',
        'height_m': 'h [m]',
        'g_mugal': 'g [µGal]',
        'sd_g_mugal': 'SD [µGal]',
        'vg_mugalm': 'VG [µGal/m]',
        'is_observed': 'Observed',
        'source_type': 'Source type',
        'source_name': ' Source name',
        'is_datum': 'Datum',
        'in_survey': 'Surveys',
    }
    _SHOW_COLUMNS_IN_TABLE = list(_SHOW_COLUMNS_IN_TABLE_DICT.keys())  # Actual list of columns to be shown

    # List of columns that are shown in the simple mode of the GUI
    _SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI = [
        'station_name',
        'long_deg',
        'lat_deg',
        'height_m',
        'g_mugal',
        'sd_g_mugal',
        'vg_mugalm',
        'is_datum',
        'is_observed',
    ]

    # Number of decimal pla
    _DECIMAL_PLACES_PER_FLOAT_COLUMN = {
        'long_deg': 8,
        'lat_deg': 8,
        'height_m': 3,
        'g_mugal': 1,
        'sd_g_mugal': 1,
        'vg_mugalm': 1,
    }

    def __init__(self, stat_df, gui_simple_mode=False):
        QAbstractTableModel.__init__(self)
        self._data_column_names = None
        self._data = None
        self.flag_gui_simple_mode = gui_simple_mode
        self.load_stat_df(stat_df)

    def load_stat_df(self, stat_df):
        """Load station data from pandas dataframe to table model."""
        # self._data = stat_df.copy()
        # Get list of columns to be depicted via the table model:
        # - Keep order of items in `self._SHOW_COLUMNS_IN_TABLE`
        stat_df_columns_set = frozenset(stat_df.columns)
        table_model_columns = [x for x in self.get_table_columns if x in stat_df_columns_set]
        # if self.flag_gui_simple_mode:
        #     self._data = stat_df.loc[:, table_model_columns]
        # else:
        #     self._data = stat_df.loc[:, table_model_columns]
        self._data = stat_df.loc[:, table_model_columns]
        self._data_column_names = self._data.columns.to_list()

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parent=None):
        return self._data.shape[1]

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():

            value = self._data.iloc[index.row(), index.column()]
            column_name = self._data_column_names[index.column()]

            if role == Qt.DisplayRole or role == Qt.UserRole:

                # Draw checkboxes only in the "is_datum" column:
                if role == Qt.DisplayRole:
                    if column_name == 'is_datum':
                        return ''

                # Custom formatter (string is expected as return type):
                if value is None:  #
                    return NONE_REPRESENTATION_IN_TABLE_VIEW
                elif isinstance(value, float):
                    if value != value:  # True, if value is "NaN"
                        return NONE_REPRESENTATION_IN_TABLE_VIEW
                    else:
                        if column_name in self._DECIMAL_PLACES_PER_FLOAT_COLUMN.keys():
                            num_dec_places = self._DECIMAL_PLACES_PER_FLOAT_COLUMN[column_name]
                            return '{1:.{0}f}'.format(num_dec_places, value)
                        else:
                            return str(value)
                else:  # all other
                    return str(value)

            if role == Qt.TextAlignmentRole:
                # value = self._data.iloc[index.row(), index.column()]
                if isinstance(value, int) or isinstance(value, float):
                    # Align right, vertical middle.
                    return Qt.AlignVCenter + Qt.AlignRight

            if role == Qt.BackgroundRole:
                if column_name == 'is_datum':
                    if value:
                        return QtGui.QColor(settings.DATUM_STATION_COLOR[0], settings.DATUM_STATION_COLOR[1],
                                            settings.DATUM_STATION_COLOR[2])

                if self._data.iloc[index.row(), self._data_column_names.index('is_observed')]:  # is_observed
                    return QtGui.QColor('cyan')

            if role == Qt.CheckStateRole:
                try:
                    if column_name == 'is_datum':
                        keep_obs_flag = self._data.iloc[index.row(), self._data_column_names.index('is_datum')]
                        if keep_obs_flag:
                            return Qt.Checked
                        else:
                            return Qt.Unchecked
                except Exception:
                    return None
        return None

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.DisplayRole:
            if self._data is not None:
                if orientation == Qt.Horizontal:
                    return self._SHOW_COLUMNS_IN_TABLE_DICT[str(self._data.columns[section])]
                if orientation == Qt.Vertical:
                    return str(self._data.index[section])

    def setData(self, index, value, role):
        """Example: https://www.semicolonworld.com/question/58510/how-to-display-a-pandas-data-frame-with-pyqt5"""
        if not index.isValid():
            return False
        if role == Qt.EditRole:
            # Get column and row indices for dataframe:
            row = self._data.index[index.row()]
            col = self._data.columns[index.column()]

            # Handle data type in a generic way, if required:
            # - list of available dtypes: https://stackoverflow.com/questions/37561991/what-is-dtypeo-in-pandas
            # dtype_col = self._data[col].dtype  # dtype of the dataframe columns
            # if dtype != object:  # "string"
            #     value = None if value == '' else dtype.type(value)  # changes dtype of the dataframe...

            if col == 'is_datum':
                # convert "value" (str) to bool and set itm in dataframe:
                if value == 'True':
                    self._data.at[row, col] = True
                    self.dataChanged.emit(index, index)  # Is it necessary?
                elif value == 'False':
                    self._data.at[row, col] = False
                    self.dataChanged.emit(index, index)  # Is it necessary?
                else:
                    QMessageBox.warning(self.parent(), 'Warning!',
                                        f'Input "{value}" not valid! Only "True" or "False" allowed.')
                    return False
                return True  # Data successfully set

        if role == Qt.CheckStateRole:
            row = self._data.index[index.row()]
            col = self._data.columns[index.column()]
            if col == 'is_datum':
                if value == Qt.Unchecked:
                    self._data.at[row, col] = False
                    self.dataChanged.emit(index, index)  # Update only one item
                elif value == Qt.Checked:
                    self._data.at[row, col] = True
                    self.dataChanged.emit(index, index)  # Update only one item
                else:
                    QMessageBox.warning(self.parent(), 'Warning!',
                                        f'Invalid value fpr keep observation flag: "{value}"')
                    return False

            return True  # Data successfully set

        return False

    def flags(self, index):
        """Enable editing of table items."""
        flags = super(self.__class__, self).flags(index)
        flags |= Qt.ItemIsSelectable
        flags |= Qt.ItemIsEnabled
        flags |= Qt.ItemIsDragEnabled
        flags |= Qt.ItemIsDropEnabled
        if index.column() == self._data_column_names.index('is_datum'):  # Column: "is_datum"
            # flags |= Qt.ItemIsEditable  # Use checkbox only!
            flags |= Qt.ItemIsUserCheckable
        return flags

    @property
    def get_data(self):
        return self._data

    @property
    def get_table_columns(self):
        """Returns a list with all columns names of the dataframe that should be copied to the view model."""
        if self.flag_gui_simple_mode:
            return [value for value in self._SHOW_COLUMNS_IN_TABLE if value in self._SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI]
        else:
            return self._SHOW_COLUMNS_IN_TABLE



"""
GravTools - A software toolbox for processing relative gravity surveys
======================================================================

The module *gui* represents the graphical user interface of GravTools written with PyQt.

Notes
-----

The .ui files are created using the Qt Designer. The according .py files are automatically created with the program
pyuic5 based on the .ui files and should therefore not be edited!

License, contact and authorship
-------------------------------

Copyright (C) 2021  Andreas Hellerschmied

GravTools was developed at Austria's the *Federal Office of Metrology and Surveying (BEV)*.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

:Authors:
    Andreas Hellerschmied (andreas.hellerschmied@bev.gv.at)
"""



# -*- coding: utf-8 -*-

from datetime import datetime
from math import radians
from typing import Union
import numpy as np
import pandas as pd


__all__ = ['calculate_julian_century', 'solve_longman_tide', 'solve_longman_tide_scalar',
           'solve_tide_df', 'solve_point_corr']


"""
Longman Earth Tide Calculator - Adapted from John Leeman's implementation of 
I. M. Longman's earth tide calculations (see references below) 

Parts of this program (c) 2017 John Leeman
Any other modifications from the base LongmanTide (c) 2018 Zachery Brady

Licensed under the MIT License, see LICENSE text below

References
----------
I.M. Longman "Formulas for Computing the Tidal Accelerations Due to the Moon 
and the Sun" Journal of Geophysical Research, vol. 64, no. 12, 1959, pp. 2351-2355
P. Schureman "Manual of harmonic analysis and prediction of tides" U.S. Coast and Geodetic Survey, 1958
John Leeman's GitHub page for the original implementation of this library: https://github.com/jrleeman/LongmanTide

Notes
-----
Unicode greek symbols are used to more clearly name variables based on the equations in Longman's paper.
This is simply a style decision and may not reflect Python best practices.
Pythons Datetime.datetime objects are TimeZone naive - e.g. when creating a datetime of
>>> datetime(1899, 12, 31, 12, 0, 0)
The object refers to exactly the time specified, with no knowledge of the time-zone - when doing calculations against
such an object we need to make sure that whatever datetime being used is specified in the same time zone.

LICENSE
-------
MIT License

Copyright (c) 2017 John Leeman
Copyright (c) 2018-2020 Zachery Brady

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

# Constants Definitions #
μ = 6.673e-8  # Newton's gravitational constant in cgs units (Verify this, should be 6.674e-8?)
# μ = 6.67408e-8  # Newton's gravitational constant in cgs units (from NIST.gov)
M = 7.3537e25  # Mass of the moon in grams
S = 1.993e33  # Mass of the sun in grams
e = 0.05490  # Eccentricity of the moon's orbit
m = 0.074804  # Ratio of mean motion of the sun to that of the moon
c = 3.84402e10  # Mean distance between the centers of the earth and the moon in cm
c1 = 1.495e13  # Mean distance between centers of the earth and sun in cm
h2 = 0.612  # Love parameter  # See: https://en.wikipedia.org/wiki/Love_number
k2 = 0.303  # Love parameter  # See Also: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4599447/
love = 1 + h2 - 1.5 * k2
a = 6.378270e8  # Earth's equatorial radius in cm
i = 0.08979719  # (i) Inclination of the moon's orbit to the ecliptic
ω = radians(23.452)  # Inclination of the Earth's equator to the ecliptic
origin_date = datetime(1899, 12, 31, 12, 00, 00)  # Noon Dec 31, 1899
# End Constants Definitions #


def calculate_julian_century(dates: Union[np.ndarray, pd.DatetimeIndex]):
    """
    Calculate the decimal Julian century and floating point hour, as referenced from
    1899, December 31 at 12:00:00
    This function accepts either a numpy ndarray or pandas DatetimeIndex as input,
    and returns a 2-tuple of the corresponding Julian century decimals and floating
    point hours.

    Parameters
    ----------
    dates : Union[np.ndarray, pd.DatetimeIndex]
        1-dimensional array of DateTime objects to convert into
        century/hours arrays

    Notes
    -----
    All DateTimes should be supplied as UTC values, using a timezone-naive DateTime object.
    This can be accomplished either by using datetime.utcnow(), or by constructing datetime
    objects where the times are specified as UTC values
    Reference Date: 1899 December 31 12:00:00
    Reference ordinal: 693961.5 (MATLAB Serial date from January 0, 0000) Delta 366 days
    Reference ordinal: 694327.5 (Python ordinal from January 1, 0001)

    Returns
    -------
    2-tuple of:
        T : np.ndarray
            Number of Julian centuries (36525 days) from GMT Noon on
            December 31, 1899
        t0 : np.ndarray
            Greenwich civil dates measured in hours
    """
    if isinstance(dates, np.ndarray):
        delta = dates - origin_date
        days = np.array([x.days + x.seconds / 3600. / 24. for x in delta])
        t0 = np.array([x.hour + x.minute / 60. + x.second / 3600. for x in dates])
        return days / 36525, t0
    elif isinstance(dates, pd.DatetimeIndex):
        delta = dates - origin_date
        days = delta.days + delta.seconds / 3600. / 24.
        t0 = dates.hour + dates.minute / 60. + dates.second / 3600.
        return days / 36525, t0.values


def solve_longman_tide(lat: np.ndarray, lon: np.ndarray, alt: np.ndarray, time: np.ndarray):
    """
    Find the total gravity correction due to the Sun/Moon for the given
    latitude, longitude, altitude, and time - supplied as numpy 1-d arrays.
    Corrections are calculated for each corresponding set of data in the
    supplied arrays, all arrays must be of the same shape (length and dimension)
    This function returns the Lunar, Solar, and Total gravity corrections as
    a 3-tuple of numpy 1-d arrays.

    Parameters
    ----------
    lat : np.ndarray
        1-dimensional array of float values denoting latitudes
    lon : np.ndarray
        1-dimensional array of float values denoting longitudes
    alt : np.ndarray
        1-dimensional array of float values denoting altitude in meters
    time : np.ndarray
        1-dimensional array of DateTime objects denoting the time series

    Returns
    -------
    3-Tuple
        gMoon (gm): Vertical component of tidal acceleration due to the moon
        gSun (gs): Vertical component of tidal acceleration due to the sun
        gTotal (g0): Total vertical component of tidal acceleration (moon + sun)
    """
    assert lat.shape == lon.shape == alt.shape == time.shape

    T, t0 = calculate_julian_century(time)
    T2 = T ** 2
    T3 = T ** 3

    # t0 must be an ndarray, not a pandas DatetimeIndex
    t0[t0 < 0] += 24
    t0[t0 >= 24] -= 24

    # longitude is defined with West positive (+) in the Longman paper
    φ = -lon
    λ = np.radians(lat)  # λ Latitude of point P
    cosλ = np.cos(λ)
    sinλ = np.sin(λ)
    H = alt * 100  # height above sea-level of point P in centimeters (cm)

    # Lunar Calculations

    # s Mean longitude of moon in its orbit reckoned from the referred equinox
    # Constants from Bartels [1957 pp. 747] eq (10')
    # 270°26'11.72" + (1336 rev. + 1,108,406.05")T + 7.128" * T2 + 0.0072" * T3
    # Where rev. is revolutions expressed as radians: 1336 rev. = 1336 * 2 * pi = 8394.335570392
    s = 4.72000889397 + 8399.70927456 * T + 3.45575191895e-05 * T2 + 3.49065850399e-08 * T3
    # p Mean longitude of lunar perigee
    # constants from Bartels [1957] eq (11')
    # p = 334° 19' 46.42" + (11 rev. + 392,522.51") T - 37.15" * T2 - 0.036" T3
    p = 5.83515162814 + 71.0180412089 * T + 0.000180108282532 * T2 + 1.74532925199e-07 * T3
    # (h) Mean longitude of the sun
    h = 4.88162798259 + 628.331950894 * T + 5.23598775598e-06 * T2
    # (N) Longitude of the moon's ascending node in its orbit reckoned from the referred equinox
    N = 4.52360161181 - 33.757146295 * T + 3.6264063347e-05 * T2 + 3.39369576777e-08 * T3
    cosN = np.cos(N)
    sinN = np.sin(N)

    # I (uppercase i) Inclination of the moon's orbit to the equator
    I = np.arccos(np.cos(ω) * np.cos(i) - np.sin(ω) * np.sin(i) * cosN)
    # ν (nu) Longitude in the celestial equator of its intersection A with the moon's orbit
    ν = np.arcsin(np.sin(i) * sinN / np.sin(I))
    # t Hour angle of mean sun measured west-ward from the place of observations
    t = np.radians(15. * (t0 - 12) - φ)

    # χ (chi) right ascension of meridian of place of observations reckoned from A
    χ = t + h - ν
    # cos α (alpha) where α is defined in eq. 15 and 16
    cos_α = cosN * np.cos(ν) + sinN * np.sin(ν) * np.cos(ω)
    # sin α (alpha) where α is defined in eq. 15 and 16
    sin_α = np.sin(ω) * sinN / np.sin(I)
    # (α) α is defined in eq. 15 and 16
    α = 2 * np.arctan(sin_α / (1 + cos_α))
    # ξ (xi) Longitude in the moon's orbit of its ascending intersection with the celestial equator
    ξ = N - α

    # σ (sigma) Mean longitude of moon in radians in its orbit reckoned from A
    σ = s - ξ
    # l (lowercase el) Longitude of moon in its orbit reckoned from its ascending intersection with the equator
    l = σ + 2 * e * np.sin(s - p) + (5. / 4) * e * e * np.sin(2 * (s - p)) + (15. / 4) * m * e * np.sin(s - 2 * h + p)\
        + (11. / 8) * m * m * np.sin(2 * (s - h))

    # Solar Calculations

    # p1 (p-one) Longitude of solar perigee (Schureman [1941, pp. 162])
    # p1 = 281° 13' 15.0" + 6189.03" T + 1.63" T2 + 0.012" T3
    p1 = 4.90822941839 + 0.0300025492114 * T + 7.85398163397e-06 * T2 + 5.3329504922e-08 * T3
    # e1 (e-one) Eccentricity of the Earth's orbit
    e1 = 0.01675104 - 0.00004180 * T - 0.000000126 * T2
    # χ1 (chi-one) right ascension of meridian of place of observations reckoned from the vernal equinox
    χ1 = t + h
    # l1 (lowercase-el(L) one) Longitude of sun in the ecliptic reckoned from the vernal equinox
    l1 = h + 2 * e1 * np.sin(h - p1)
    # cosθ (theta) θ represents the zenith angle of the moon
    cosθ = sinλ * np.sin(I) * np.sin(l) + cosλ * (np.cos(0.5 * I) ** 2 * np.cos(l - χ) + np.sin(0.5 * I) ** 2 *
                                                  np.cos(l + χ))
    # cosφ (phi) φ represents the zenith angle of the run
    cosφ = sinλ * np.sin(ω) * np.sin(l1) + cosλ * (np.cos(0.5 * ω) ** 2 * np.cos(l1 - χ1) + np.sin(0.5 * ω) ** 2 *
                                                np.cos(l1 + χ1))

    # Distance Calculations

    # (C) Distance parameter, equation 34
    # C**2 = 1/(1 + 0.006738 sinλ ** 2)
    C = np.sqrt(1. / (1 + 0.006738 * sinλ ** 2))
    # (r) Distance from point P to the center of the Earth
    r = C * a + H
    # a' (a prime) Distance parameter, equation 31
    aprime = 1. / (c * (1 - e * e))
    # a1' (a-one prime) Distance parameter, equation 31
    aprime1 = 1. / (c1 * (1 - e1 * e1))
    # (d) Distance between centers of the Earth and the moon
    d = 1. / ((1. / c) + aprime * e * np.cos(s - p) + aprime * e ** 2 * np.cos(2 * (s - p)) + (15. / 8) * aprime * m *
              e * np.cos(s - 2 * h + p) + aprime * m * m * np.cos(2 * (s - h)))
    # (D) Distance between centers of the Earth and the sun
    D = 1. / ((1. / c1) + aprime1 * e1 * np.cos(h - p1))

    # (gm) Vertical component of tidal acceleration due to the moon, equation (1):
    # gm = μMr/d^3 (3 * cos^2(θ) - 1) + 3/2 μMr/d^4 * (5 cos^3 θ - 3 cos(θ))
    gm = (μ * M * r / d ** 3) * (3 * cosθ ** 2 - 1) + (1.5 * (μ * M * r ** 2 / d ** 4) * (5 * cosθ ** 3 - 3 * cosθ))
    # (gs) Vertical component of tidal acceleration due to the sun
    gs = μ * S * r / D ** 3 * (3 * cosφ ** 2 - 1)

    # g0 Total vertical component due to Lunar and Solar forces
    g0 = (gm + gs) * 1e3 * love

    # Returns Lunar, Solar, Total corrections in mGals, as numpy 1d-arrays
    return gm * 1e3 * love, gs * 1e3 * love, g0


def solve_longman_tide_scalar(lat: float, lon: float, alt: float, time: datetime):
    """
    Simple wrapper around solve_longman_tide that allows passing of singular scalar values instead of an array.
    This function simply wraps the scalar values within an ndarray and then extracts the result elements.
    Parameters
    ----------
    lat : float
        Latitude as a scalar float value
    lon : float
        Longitude as a scalar float value
    alt : float
        Altitude as a scalar float value
    time : datetime
        Time as a scalar datetime object
    Returns
    -------
    gm, gs, g0 : float, float, float
        Where gm is the gravitational effect due to the moon
              gs is the gravitational effect due to the sun
              g0 is the total gravitational effect of the sun and moon (gm+gs)
    """
    lat_arr = np.array([lat])
    lon_arr = np.array([lon])
    alt_arr = np.array([alt])
    time_arr = np.array([time])

    gm, gs, g0 = solve_longman_tide(lat_arr, lon_arr, alt_arr, time_arr)
    return gm[0], gs[0], g0[0]


def solve_point_corr(lat: float, lon: float, alt: float, t0=datetime.utcnow(), n=3600, increment='S'):
    """
    Utility function to generate a tide correction DataFrame for a static lat/lon/alt given start time t0,
    an increment, and count (n) of datapoints to generate.
    Default parameters are supplied that will generate a correction series with one second increment over a one hour
    period, with start time being the time of execution.

    Parameters
    ----------
    lat : float
        Latitude in decimal degrees
    lon : float
        Longitude in decimal degrees
    alt : float
        Altitude (height) above sea level in meters
    t0 : DateTime
        Starting date/time
    n : int
        Number of data points to generate
    increment : String
        Increment between data points, uses Pandas offset aliases:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timedelta.html#pandas.Timedelta

        Common options:
            * H : Hourly Frequency
            * T, min : Minutely frequency
            * S : Secondly frequency
            * L, ms : millisecond frequency
            * U, us : microsecond frequency
            * N : nanosecond frequency

    Returns
    -------
    df : pd.DataFrame
        DataFrame (index=DateTime, lat, lon, alt, gm, gs, g0) shape: (n, 4)
        Pandas DataFrame indexed by DateTime containing the latitude, longitude, altitude, lunar, solar,
        and total corrections.
    Notes
    -----
    total_corr column renamed to g0 and gm/gs added in v0.3.0
    """
    df = pd.DataFrame(data={'lat': np.repeat(lat, n), 'lon': np.repeat(lon, n), 'alt': np.repeat(alt, n)},
                      index=pd.date_range(start=t0, freq=increment, periods=n))

    gm, gs, g0 = solve_longman_tide(df.lat, df.lon, df.alt, df.index)
    df['gm'] = gm
    df['gs'] = gs
    df['g0'] = g0
    return df


def solve_tide_df(df: pd.DataFrame, lat='lat', lon='lon', alt='alt'):
    """
    Solve tidal gravity corrections for a given Pandas DataFrame, returning the source
    DataFrame with an appended 'tide_corr' column.
    Source DataFrame should be indexed by datetime and must contain latitude, longitude,
    and altitude columns.
    Source column names can be specified by passing the applicable column name to the
    lat, lon, alt parameters as required.
    Time is assumed to be the index of the DataFrame
    Returns
    -------
    df : pd.DataFrame
        Copy of source df with added columns 'gm' 'gs' and 'g0': lunar, solar and total corrections respectively
    """
    _lat = df[lat].values
    _lon = df[lon].values
    _alt = df[alt].values
    _time = df.index
    res_df = df.copy(deep=True)  # type: pd.DataFrame

    gm, gs, g0 = solve_longman_tide(_lat, _lon, _alt, _time)
    res_df['gm'] = gm
    res_df['gs'] = gs
    res_df['g0'] = g0
    return res_df


"""Modelling of relative gravity campaigns.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import datetime as dt
import pytz
import numpy as np
import pickle
import os
import sys

import pandas as pd
from gravtools.models.lsm import LSM
from gravtools.models.lsm_diff import LSMDiff
from gravtools.models.vg_lsm import VGLSM
from gravtools.models.lsm_nondiff import LSMNonDiff
from gravtools.models.mlr_bev_legacy import BEVLegacyProcessing
from gravtools.models.survey import Survey
from gravtools.models.station import Station
from gravtools.models.gravimeter import Gravimeters
from gravtools.tides.correction_time_series import CorrectionTimeSeries
from gravtools.settings import ADDITIVE_CONST_ABS_GRAVITY, EXPORT_OBS_LIST_COLUMNS, \
    MAX_SD_FOR_EXPORT_TO_NSB_FILE, WRITE_COMMENT_TO_NSB, PICKLE_PROTOCOL_VERSION
from gravtools import __version__ as GRAVTOOLS_VERSION


class Campaign:
    """Gravity Campaign dataset.

    A gravity campaign datasets consists of:

    - Campaign name
    - One or more gravity surveys that belong together and
      - Each survey was observed with one gravimeter on a single day
    - Station data (datum and non-datum stations)
    - Reductions and corrections
      - All observations (from surveys) are corrected and reduced in the same way
    - Time series data for the correction of gravity observations (optional)
    - Gravimeter data

    Attributes
    ----------
    campaign_name : str
        Name of the campaign.
    output_directory : str
        Path to output directory (all output files are stored there).
    surveys: dict of :py:obj:`.Survey` objects
        Arbitrary number of survey objects.
    stations : :py:obj:`.Station` object
        Data of known stations (datum- and non-datum-stations).
    lsm_runs : list of objects inherited from :py:obj:`gravtools.models.lsm.LSM`
        Each item in the list contains one enclosed LSM object. Each LSM object reflects one dedicated run of a
        least-squares adjustment in order to estimate target parameters.
    ref_delta_t_dt : datetime object
        Reference epoch for relative times within the campaign, e.g. for the determination of the drift polynomials.
        This reference time is equalt to the first (active) observation in the campaign considering all surveys.
    correction_time_series : :py:obj:`CorrectionTimeSeries`
        Contains time series data for correcting gravity observations.
    gravimeters : :py:obj:`Gravimeters`

    gravtools_version : str
        Version of the gravtools software that was used to create the dataset.
    """

    def __init__(self,
                 campaign_name,
                 output_directory,
                 surveys=None,  # Always use non-mutable default arguments!
                 stations=None,  # Always use non-mutable default arguments!
                 gravimeters=None,  # Always use non-mutable default arguments!
                 lsm_runs=None,  # Always use non-mutable default arguments!
                 ref_delta_t_dt=None  # Reference time for drift determination
                 ):
        """
        Parameters
        ----------
        campaign_name : str
            Name of the campaign.
        surveys: dict of :py:obj:`.Survey` objects, optional
            Arbitrary number of survey data objects. Default=None which implies that the campaign will be initialized
            without surveys.
        stations: :py:obj:`.Station` object, optional
            Station data (datum- and non-datum-stations). Default=None implies that the campaign will be
            initialized without station data.
        gravimeters: :py:obj:`.Gravimeters` object, optional
            Gravimeter data. Default=None implies that the campaign will be
            initialized without gravimeter data.
        lsm_runs : list of objects inherited from :py:obj:`gravtools.models.lsm.LSM`
            Each item in the list contains one enclosed LSM object. Each LSM object reflects one dedicated run of an
            least-squares adjustment in order to estimate target parameters.

        Raises
        ------
        TypeError
            Wrong input argument type.
        """

        # Check campaign_name:
        if not isinstance(campaign_name, str):
            raise TypeError('The argument "campaign_name" needs to be a string.')
        else:
            if not campaign_name:
                raise ValueError('"campaign_name" should not be empty!')
        self.campaign_name = campaign_name

        # Check output directory:
        if not isinstance(output_directory, str):
            raise TypeError('The argument "output_directory" needs to be a string.')
        else:
            if not output_directory:
                raise ValueError('"output_directory" should not be empty!')
        self.output_directory = output_directory

        # Check surveys:
        if surveys is None:
            surveys = {}
        else:
            if not isinstance(surveys, dict):
                raise TypeError('The argument "survey" needs to be a dict of Survey objects.')
            else:
                for survey_name, survey_obj in surveys.items():
                    if not isinstance(survey_name, str):
                        raise TypeError('The argument "survey" needs to be a string.')
        self.surveys = surveys  # dict: key=Name of Survey, value=Survey object

        # Check stations:
        if stations is None:
            stations = Station()
        else:
            if not isinstance(stations, Station):
                raise TypeError('The argument "stations" needs to be a Station object.')
        self.stations = stations

        # Check gravimeters:
        if gravimeters is None:
            gravimeters = Gravimeters()
        else:
            if not isinstance(gravimeters, Gravimeters):
                raise TypeError('The argument "gravimeters" needs to be a Gravimeters object.')
        self.gravimeters = gravimeters

        # Check lsm_runs:
        if lsm_runs is None:
            lsm_runs = []  # Empty list
        else:
            if not isinstance(lsm_runs, list):
                raise TypeError('The argument "lsm_runs" needs to be a list of LSM-objects.')
            else:
                for lsm_run in lsm_runs:
                    if not isinstance(lsm_run, LSM):
                        raise TypeError('The argument "lsm_runs" needs to be a list of LSM-objects.')
        self.lsm_runs = lsm_runs

        # Check ref_delta_t_dt:
        if ref_delta_t_dt is not None:
            if not isinstance(ref_delta_t_dt, dt.datetime):
                raise TypeError('`ref_delta_t_dt` needs to be a datetime object.')
        self.ref_delta_t_dt = ref_delta_t_dt

        # Version of gravtools:
        self.gravtools_version = GRAVTOOLS_VERSION

        # Correction time series object:
        self.correction_time_series = CorrectionTimeSeries()

    def add_empty_correction_time_series(self):
        """Adds an empty `CorrectionTimeSeries` object to the campaign.

        Notes
        -----
        This is required, e.g. if a campaign object os loaded into GravTools from a previous GRavTools version without
        support of time series corrections.
        """
        self.correction_time_series = CorrectionTimeSeries()

    def init_gravimeters(self):
        """Adds an empty `Gravimeters` object to the campaign.

        Notes
        -----
        This is required, e.g. if a campaign object os loaded into GravTools from a previous GravTools version without
        support of gravimeter data.
        """
        self.gravimeters = Gravimeters()

    def add_survey(self, survey_add: Survey, verbose=False):
        """Add a survey to campaign and specify whether to use it for ths analysis.

        Notes
        -----
        A survey can only be added, f the survey's name is unique within the campaign.

        Parameters
        ----------
        survey_add : :py:obj:`.Survey`
            Contains all information of s specific survey independent of the data source.
        verbose : bool, optional (default=False)
            If True, status messages are printed to the command line.
        """
        # Check if a survey with the dame name ("survey_add.name") already exists in this campaign:
        # - Raise warning:
        if survey_add.name in self.surveys.keys():
            raise RuntimeError(f'The campaign already contains a survey named {survey_add.name}. Survey names need to '
                               f'be unique within a campaign!')
        else:
            # Add survey:
            self.surveys[survey_add.name] = survey_add
            if verbose:
                print(f"Survey {survey_add.name} added to the campaign.")

    def remove_survey(self, survey_name: str, verbose=False) -> bool:
        """Remove survey with the specified name from the campaign

        Parameters
        ----------
        survey_name : str
            Name of the survey that will be removed from the campaign.
        verbose : bool, optional (default=False)
            If True, status messages are printed to the command line.

        Returns
        -------
        bool
            True, if the survey was successfully removed; False, if not.
        """
        try:
            del self.surveys[survey_name]
            if verbose:
                print(f'Survey "{survey_name}" removed from campaign.')
        except KeyError:
            if verbose:
                print(f'Survey "{survey_name}" does not exist.')
            return False
        except Exception:
            if verbose:
                print(f'Failed to remove survey {survey_name}.')
            return False
        else:
            return True

    def activate_survey(self, survey_name: str, verbose=False) -> bool:
        """Set the survey with the specified name active.

        Parameters
        ----------
        survey_name : str
            Name of the survey that will be set active.
        verbose : bool, optional (default=False)
            If True, status messages are printed to the command line.

        Returns
        -------
        bool
            True, if the survey was successfully activated; False, if not.
        """
        try:
            if self.surveys[survey_name].keep_survey:
                if verbose:
                    print(f'Survey "{survey_name}" already active.')
                return True
            else:
                self.surveys[survey_name].keep_survey = True
                if verbose:
                    print(f'Survey "{survey_name}" activated.')
        except KeyError:
            if verbose:
                print(f'Survey "{survey_name}" does not exist.')
            return False
        except:
            if verbose:
                print(f'Failed to activate survey "{survey_name}"')
            return False
        else:
            return True

    def deactivate_survey(self, survey_name: str, verbose=False) -> bool:
        """Set the survey with the specified name inactive.

        Parameters
        ----------
        survey_name : str
            Name of the survey that will be set inactive.
        verbose : bool, optional (default=False)
                If True, status messages are printed to the command line.

        Returns
        -------
        bool
            True, if the survey was successfully deactivated; False, if not.
        """
        try:
            if not self.surveys[survey_name].keep_survey:
                if verbose:
                    print(f'Survey "{survey_name}" already inactive.')
                return True
            else:
                self.surveys[survey_name].keep_survey = False
                if verbose:
                    print(f'Survey "{survey_name}" deactivated.')
        except KeyError:
            if verbose:
                print(f'Survey "{survey_name}" does not exist.')
            return False
        except:
            if verbose:
                print(f'Failed to deactivate survey "{survey_name}"')
            return False
        else:
            return True

    def deactivate_all_surveys(self, verbose=False):
        """Deactivates all surveys in the campaign.

        Parameters
        ----------
        verbose : bool, optional (default=False)
                If True, status messages are printed to the command line.
        """
        changed_surveys = []
        for survey_name, survey in self.surveys.items():
            if self.deactivate_survey(survey_name=survey_name, verbose=verbose):
                changed_surveys.append(survey_name)
        return changed_surveys

    def activate_all_surveys(self, verbose=False):
        """Activates all surveys in the campaign.

        Parameters
        ----------
        verbose : bool, optional (default=False)
                If True, status messages are printed to the command line.
        """
        changed_surveys = []
        for survey_name, survey in self.surveys.items():
            if self.activate_survey(survey_name=survey_name, verbose=verbose):
                changed_surveys.append(survey_name)
        return changed_surveys

    def get_survey_names_and_status(self, verbose: bool = False) -> dict:
        """Return list with all survey names and information whether the survey is set active.

        Parameters
        ----------
        verbose : bool, optional (default=False)
            If True, survey names and status are printed to the command line.

        Returns
        -------
        dict
            The keys are the survey names and the values represent the respective status (active=True, inactive=False).
        """
        if verbose:
            print('Surveys and their status:')
        info_dict = {}
        lookup_dict = {True: 'active', False: 'inactive'}
        for surv_name, surv_obj in self.surveys.items():
            if verbose:
                activity_str = lookup_dict[surv_obj.keep_survey]
                print(f' - {surv_name:12s} ({activity_str:8s}): {surv_obj.get_number_of_observations()} observations')
            info_dict[surv_name] = surv_obj.keep_survey
        return info_dict

    @property
    def survey_names(self):
        """Returns a list with the names of all surveys in the campaign."""
        return list(self.surveys.keys())

    @property
    def number_of_surveys(self) -> int:
        """int : Returns the number of surveys in this campaign."""
        return len(self.surveys)

    @property
    def number_of_stations(self) -> int:
        """int : Returns the number of stations in this campaign."""
        return self.stations.get_number_of_stations

    def reduce_observations_in_all_surveys(self,
                                           target_ref_height=None,
                                           target_tide_corr=None,
                                           target_atm_pres_corr=None,
                                           target_scale_corr=None,
                                           target_oceanload_corr=None,
                                           atm_pres_admittance=None,
                                           tide_corr_timeseries_interpol_method='',
                                           verbose=False):
        """Reduce the observed gravity by applying the specified corrections.

        Notes
        -----
        - For this reduction vertical gravity gradients are required. They are obtained from the `Station` object.
          Hence, a `Station` object has to be attached to the Campaign object beforehand.

        - All corrections are applied on the survey-level. See py:obj:`.Survey.reduce_observations` for more details.

        Parameters
        ----------
        target_ref_height : string, specifying the target reference height type (default = `None`).
            The target reference height type has to be listed in :py:obj:`gravtools.settings.REFERENCE_HEIGHT_TYPE`.
            Default is `None` indicating that the reference heights of the input data are not changed.
        target_tide_corr : str, specifying the tidal correction type to be applied (default = `None`).
            The target tidal correction type specifies what kind of tidal correction will be applied. Valid types have
            to be listed in :py:obj:`gravtools.settings.TIDE_CORRECTION_TYPES`. Default is `None` indicating that the
            tidal corrections are not considered here (tidal corrections are inherited from input data).
        target_atm_pres_corr : str (default = `None`)
            Specifying the atmospheric press ure correction type to be applied to all surveys. Valid types to be listed
            in :py:obj:`gravtools.settings.ATM_PRES_CORRECTION_TYPES`. Default is `None` indicating that the respective
            corrections of the input data are not changed.
        target_scale_corr : str, optional (default = `None`)
            Specifies the scale correction type to be applied on all observations of this survey. Vali types are listed
            in :py:obj:`gravtools.settings.SCALE_CORRECTION_TYPES`. The default value `None` indicates that the
            respective corrections of the input data will not be changed.
        target_oceanload_corr : str, optional (default = `None`)
            Specifying the ocean-loading correction type to be applied to all surveys (if available). Valid types to be
            listed in :py:obj:`gravtools.settings.OCEANLOAD_CORRECTION_TYPES`. Default is `None` indicating that the
            respective corrections of the input data are not changed.
        atm_pres_admittance : float, optional (default = `None`)
            Admittance factor for the determination of pressure corrections based on the difference between measured and
            normal air pressure. If `target_atm_pres_corr` is not None (i.e. atmospheric pressure corrections will be
            calculated), `atm_pres_admittance` has to be provided too (float). Otherwise, an error is raised.
        tide_corr_timeseries_interpol_method : str, optional (default='')
            Interpolation method used to calculate tidal corrections from time series data. If tidal corrections are
            obtained from other sources or models, this attribute is irrelevant and has to be empty!
        verbose : bool, optional (default=False)
            If True, status messages are printed to the command line.
        """
        if verbose:
            print(f'## Reduce all observation in this campaign:')
        for survey_name, survey in self.surveys.items():
            if verbose:
                print(f'Survey {survey_name}:')
            if target_ref_height is not None:
                if verbose:
                    print(f' - Get vertical gradients')
                survey.obs_df_populate_vg_from_stations(self.stations, verbose=verbose)
            survey.obs_df_populate_locations_from_stations(self.stations, verbose=verbose)
            survey.reduce_observations(
                target_ref_height=target_ref_height,
                target_tide_corr=target_tide_corr,
                target_atm_pres_corr=target_atm_pres_corr,
                target_scale_corr=target_scale_corr,
                atm_pres_admittance=atm_pres_admittance,
                target_oceanload_corr=target_oceanload_corr,
                tide_corr_timeseries_interpol_method=tide_corr_timeseries_interpol_method,
                correction_time_series=self.correction_time_series,
                gravimeters=self.gravimeters,
                verbose=verbose)

    def add_stations_from_oesgn_table_file(self, oesgn_filename, is_datum=False, verbose=False):
        """Add station from an OESGN table file.

        Parameters
        ----------
        oesgn_filename : string, specifying the path and filename of the OESGN file
            Stations in the specified OESGN table file are added to the campaign.
        is_datum : bool, optional (default = False)
            `True` indicates that all loaded OESGN stations are initially selected as datum stations (is_datum=True)
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        self.stations.add_stations_from_oesgn_table(filename=oesgn_filename, is_datum=is_datum, verbose=verbose)

    def add_stations_from_csv_file(self, csv_filename, verbose=False):
        """Add station from a CSV file.

        Parameters
        ----------
        csv_filename : string, specifying the path and filename of the station csv file
            Stations in this csv file are added to the campaign.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        self.stations.add_stations_from_csv_file(filename=csv_filename, verbose=verbose)

    def synchronize_stations_and_surveys(self, verbose=False):
        """Synchronize information between station and survey data in the campaign.

        The following information is synchronized:

        - The `is_observed` flags in the :py:obj:`.Campaign.stations.stat_df` are set according to the surveys in
          :py:obj:`.Campaign.surveys`. `True` indicated that the station as observed at least once.
        - Populates the vertical gradient columns (``) of the observation DataFrames (:py:obj:`.Campaign.surveys`) with
          values from a Station object (:py:obj:`.Campaign.stations`).
        - Populate longitudes, latitudes and heights in the observation DataFrames (:py:obj:`.Campaign.surveys`) with
          values from a Station object (:py:obj:`.Campaign.stations`).
        - Add observed stations

        Notes
        -----
        It is recommended to run this method whenever new stations and/or new surveys are added to the campaign on order
        to synchronize the data.

        Parameters
        ----------
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        self.stations.stat_df['is_observed'] = False  # Reset to default.
        self.stations.stat_df['in_survey'] = None  # Reset to default.
        self.sync_observed_stations(verbose=verbose)  # Add stations from surveys.

        # Loop over all surveys to match and synchronize the survey data with the station data:
        for survey_name, survey in self.surveys.items():
            if verbose:
                print(f' - Survey: {survey_name}')
            survey.obs_df_populate_vg_from_stations(self.stations, verbose=verbose)
            survey.obs_df_populate_locations_from_stations(self.stations, verbose=verbose)
            self.stations.set_observed_info_from_survey(survey)

    def sync_observed_stations(self, verbose=False):
        """Adds all stations that were observed in at least one survey to this campaign's station dataframe.

        Parameters
        ----------
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        # Loop over surveys in this campaign:
        for survey_name, survey in self.surveys.items():
            if verbose:
                print(f' - Survey: {survey_name}')
            self.stations.add_stations_from_survey(survey, verbose)

    def calculate_setup_data(self,
                             obs_type='reduced',
                             active_obs_only_for_ref_epoch=True,
                             method='variance_weighted_mean',
                             method_sd='sd_from_obs_file',
                             default_sd_mugal=100.0,
                             verbose=False):
        """Calculate accumulated pseudo observations for each active setup in all active surveys.

        Notes
        -----
        Two relative reference epochs are calculated for each setup: (a) w.r.t. the first (active) observation in the
        whole campaign and (b) w.r.t. the first (active) observation in each survey. Both reference time do not differ
        for the first survey on a campaign. Whether active observations only are considered is defined by the input
        parameter `active_obs_only_for_ref_epoch`.

        Parameters
        ----------
        obs_type : str, 'observed' or 'reduced' (default)
            Defines whether the observed (as loaded from an observation file) or the reduced observations from
            `self.obs_df` are used to determine the weighted mean values per setup.
        active_obs_only_for_ref_epoch: bool, optional (default=True)
            `True` implies that the relative reference epochs are determined by considering active observations only.
        method : str, optional (default=`variance_weighted_mean`)
            Select method for the calculation of setup data. `variance_weighted_mean` implies that setup observations
            (observed gravity, standard deviations and reference time) are calculated by variance weighted mean of the
            individual observations. `individual_obs` implies that the original observations are used as setup data
            without any aggregation.
        method_sd : str, optional (default='sd_from_obs_file')
            Method for the determination of standard deviations (SD) of setup observations. `sd_from_obs_file` implies
            that SD are taken from the observation file. `sd_default_per_obs` and `sd_default_per_setup` imply that the
            given default SD is used, where the default SD is applied the individual observations in the first case and
            to setups in the second case. If applied to observations, the number of observations per setup still plays a
            role for weighting the setup observations in the adjustment.
        default_sd_mugal : float, optional (default=100.0)
            Default standard deviation [µGal] that is used to determine the SD of setup observations when `method_sd` is
            `sd_default_per_obs` or `sd_default_per_setup`
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        # Get reference epoch:
        self.ref_delta_t_dt = self.get_epoch_of_first_observation(active_obs_only_for_ref_epoch)

        # Loop over all surveys in the campaign:
        if verbose:
            print(f'Calculate setup data:')
        for survey_name, survey in self.surveys.items():
            if verbose:
                print(f' - Survey: {survey_name}')
            if survey.is_active and survey.keep_survey:
                survey.calculate_setup_data(gravimeters=self.gravimeters,
                                            obs_type=obs_type,
                                            ref_delta_t_campaign_dt=self.ref_delta_t_dt,
                                            active_obs_only_for_ref_epoch=active_obs_only_for_ref_epoch,
                                            method=method,
                                            method_sd=method_sd,
                                            default_sd_mugal=default_sd_mugal,
                                            verbose=verbose)
            else:
                survey.reset_setup_data(verbose)  # Remove setup data from previous calculations

    def get_epoch_of_first_observation(self, active_obs_only_for_ref_epoch=True):
        """Returns the epoch of the first (active) observation in this campaign.

        Parameters
        ----------
        active_obs_only_for_ref_epoch: bool, optional (default=True)
            `True` implies that the reference epoch is determined by considering active observations only.

        Returns
        -------
        datetime object
        """
        first_obs_epoch_dt = None
        flag_first_survey_in_campaign = True

        for survey_name, survey in self.surveys.items():

            # Set filter to select active observations only:
            if active_obs_only_for_ref_epoch:
                filter_tmp = survey.obs_df['keep_obs'] == True  # Select active observations only
            else:
                filter_tmp = [True] * len(survey.obs_df)  # Select all observations

            if len(survey.obs_df.loc[filter_tmp, 'obs_epoch']) > 0:
                if flag_first_survey_in_campaign:
                    flag_first_survey_in_campaign = False
                    first_obs_epoch_dt = survey.obs_df.loc[filter_tmp, 'obs_epoch'].min()
                else:
                    if survey.obs_df.loc[filter_tmp, 'obs_epoch'].min() < first_obs_epoch_dt:
                        first_obs_epoch_dt = survey.obs_df.loc[filter_tmp, 'obs_epoch'].min()
        return first_obs_epoch_dt

    def initialize_and_add_lsm_run(self, lsm_method, comment='', write_log=True):
        """Initialize and add an least-squares adjustment run (object) to the campaign.

        Parameters
        ----------
        lsm_method : str
            Defines the adjustment method. Has to b listed in :py:obj:`gravtools.settings.ADJUSTMENT_METHODS`.
        comment : str, optional (default = '')
            Optional comment on the adjustment run.
        write_log : bool, optional (default=True)
            Flag that indicates whether log string should be written or not.
        """
        # Initialize LSM object:
        if lsm_method == 'LSM_diff':
            lsm_run = LSMDiff.from_campaign(self, comment, write_log)
        elif lsm_method == 'LSM_non_diff':
            lsm_run = LSMNonDiff.from_campaign(self, comment, write_log)
        elif lsm_method == 'MLR_BEV':
            lsm_run = BEVLegacyProcessing.from_campaign(self, comment, write_log)
        elif lsm_method == 'VG_LSM_nondiff':
            lsm_run = VGLSM.from_campaign(self, comment, write_log)
        else:
            raise AssertionError(f'Unknown LSM method: {lsm_method}')
        # Add LSM object to campaign:
        self.lsm_runs.append(lsm_run)

    @property
    def lsm_run_times(self):
        """Returns a list of lsm-run times/dates that can be used to identify individual runs.

        Returns
        -------
        list : List of string stating the epochs of lsm adjustment runs that can be used to identify individual runs.
        """
        lsm_run_times = []
        for lsm_run in self.lsm_runs:
            lsm_run_times.append(lsm_run.time_str)
        return lsm_run_times

    @property
    def number_of_lsm_runs(self):
        """Returns the number of LSM runs in the campaign."""
        if self.lsm_runs is None:
            return 0
        else:
            return len(self.lsm_runs)

    def delete_lsm_run(self, idx):
        """Delete the LSM run with the specified index in the list.

        Parameters
        ----------
        idx : int
            Index of the LSM object in the list.
        """
        if idx != -1:
            del self.lsm_runs[idx]

    def delete_all_lsm_runs(self):
        """Delete all LSM runs in the campaign."""
        if self.lsm_runs is not None:
            self.lsm_runs.clear()

    def set_reference_time(self, ref_delta_t_dt):
        """Set reference time for the determination of relative time spans, e.g. for the drift polynomial.

        Parameters
        ----------
        ref_delta_t_dt : datetime object
            Reference time epoch w.r.t. UTC.
        """
        if isinstance(ref_delta_t_dt, dt.datetime):
            self.ref_delta_t_dt = ref_delta_t_dt
        else:
            raise ValueError('`ref_delta_t_dt` needs to be a datetime object.')

    def write_nsb_file(self, filename_nsb: str, lsm_run_index, vertical_offset_mode: str = 'first',
                       exclude_datum_stations=False, formal_error_type='se', verbose=False):
        """Write the results of an LSM run to a nsb file (input for NSDB database).

        Notes
        -----
        A nsb file can only be written, if station results are available which is not the case e.g. for the estimation
        of vertical gravity gradients!

        Parameters
        ----------
        filename_nsb : str
            Name and path of the output nsb file (e.g. /home/johnny/example.nsb)
        lsm_run_index : int
            Index of the lsm run in `campaign.lsm_runs` of which the results are exported to the nsb file.
        vertical_offset_mode : str, optional (default='first')
            Defines how the vertical offsets between instrument top and ground (dhb) and reference marker (dhf),
            respectively, are determined in the case of multiple measurements (setups) on the same point. In the nsb
            file only one dhf/dhb pair per station is allowed. Two options: (1) 'first' indicates that dhb and dhf are
            taken from the first setup at a station. (2) 'mean' indicates that mean values over all setups are taken.
        exclude_datum_stations : boolean, optional (default=False)
            `True` indicates that datum stations are excluded from the nsb file.
        formal_error_type : str, optional (default=`se`, alternative: `sd`)
            Select which formal errors (of station gravities) are exported to the nsd file. The user may choose
            between post-fit standard deviations (`sd`) and standard errors ('se'). The standard errors consider the
            noise floor defined in the estimation settings.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        # Init.:
        nsb_string = ''
        formal_error_type_options = ('sd', 'se')

        if formal_error_type not in formal_error_type_options:
            raise AssertionError(f'Invalid formal error type! Valid: "sd" and "se".')

        # Get and prepare data:
        # - lsm_run
        lsm_run = self.lsm_runs[lsm_run_index]
        results_stat_df = lsm_run.get_results_stat_df

        # Check, if station results are available (e.g. nor the case for VG estimation):
        if results_stat_df is not None:

            # Check if the required columns are available:
            if 'g_est_mugal' in results_stat_df.columns and 'sd_g_est_mugal' in results_stat_df.columns:

                # Check if the data is suitable for export to the nsb file:
                if results_stat_df.loc[results_stat_df['sd_g_est_mugal'] > MAX_SD_FOR_EXPORT_TO_NSB_FILE,
                                       'sd_g_est_mugal'].any():
                    raise AssertionError(
                        f"The SD of at least one station's estimated gravity is larger than {MAX_SD_FOR_EXPORT_TO_NSB_FILE} µGal! ")

                # Loop over stations in results dataframe:
                for index, row in results_stat_df.iterrows():

                    # Skip datum stations:
                    if exclude_datum_stations:
                        if row['is_datum']:
                            continue

                    station_name = row['station_name']
                    observed_in_surveys = []
                    dhb_list_m = []
                    dhf_list_m = []

                    # Get surveys at which the station was observed:
                    for survey_name, setup_data in lsm_run.setups.items():
                        setup_df = setup_data['setup_df']
                        if len(setup_df.loc[setup_df['station_name'] == station_name]) > 0:  # observed in this setup!
                            observed_in_surveys.append(survey_name)
                            obs_df = self.surveys[survey_name].obs_df
                            # setup_ids = obs_df.loc[obs_df['station_name'] == station_name, 'setup_id'].unique()
                            setup_ids = setup_df.loc[setup_df['station_name'] == station_name, 'setup_id'].to_list()
                            # Get list of dhb and dhf:
                            for setup_id in setup_ids:
                                dhb_list_m.append(obs_df.loc[obs_df['setup_id'] == setup_id, 'dhb_m'].values[0])
                                dhf_list_m.append(obs_df.loc[obs_df['setup_id'] == setup_id, 'dhf_m'].values[0])

                    if vertical_offset_mode == 'first':
                        dhb_m = dhb_list_m[0]
                        dhf_m = dhf_list_m[0]
                    elif vertical_offset_mode == 'mean':
                        dhb_m = np.mean(dhb_list_m)
                        dhf_m = np.mean(dhf_list_m)

                    # Get gravimeter S/N and gravimeter type of first survey in the list:

                    if len(observed_in_surveys) > 1:
                        if verbose:
                            print(f'WARNING: station {station_name} was observed in {len(observed_in_surveys)} surveys! Hence, '
                                  f'the gravimeter serial number/type and the observation date may be ambiguous in the nsb file!')
                            print(
                                f' - {station_name} was observed the following surveys: {", ".join(observed_in_surveys)}')
                        # Get the latest survey in which the station was observed:
                        latest_survey_name = ''
                        latest_survey_date = dt.date(1900, 1, 1)
                        for survey in observed_in_surveys:
                            if self.surveys[survey].date > latest_survey_date:
                                latest_survey_date = self.surveys[survey].date
                                latest_survey_name = survey
                    else:
                        latest_survey_name = observed_in_surveys[0]
                    gravimeter_type = self.surveys[latest_survey_name].gravimeter_type
                    gravimeter_serial_number = self.surveys[latest_survey_name].gravimeter_serial_number
                    gravi = self.gravimeters.gravimeters[(gravimeter_type, gravimeter_serial_number)]
                    date_str = self.surveys[latest_survey_name].date.strftime('%Y%m%d')

                    # Comment string:
                    # - Max. 5 characters!
                    if WRITE_COMMENT_TO_NSB == 'cg5_serial_number':
                        comment_str = str(gravimeter_serial_number)
                    elif WRITE_COMMENT_TO_NSB == 'gravtools_version':
                        comment_str = 'GT'+''.join(GRAVTOOLS_VERSION.split('.'))
                    else:
                        raise AssertionError(f'Invalid choice for the nsb file comment: {WRITE_COMMENT_TO_NSB}!')

                    if formal_error_type == 'se':
                        formal_error = row['se_g_est_mugal']
                    elif formal_error_type == 'sd':
                        formal_error = row['sd_g_est_mugal']

                    nsb_string += '{:10s} {:8s}  {:9.0f} {:3.0f} {:1s}{:>5s} {:4.0f} {:4.0f}\n'.format(
                        station_name,
                        date_str,
                        row['g_est_mugal'] + ADDITIVE_CONST_ABS_GRAVITY,
                        formal_error,
                        gravi.code,
                        comment_str,
                        (dhb_m + gravi.height_offset_m) * 100,
                        (dhf_m + gravi.height_offset_m) * 100,
                    )

                # Write file:
                with open(filename_nsb, 'w') as out_file:
                    out_file.write(nsb_string)

            else:  # Required columns are not available
                if verbose:
                    print(f'The nsb file cannot be written as the required station data is not available.')

        else:  # No station results available
            if verbose:
                print(f'The nsb file cannot be written as the required station data is not available.')
            
    def write_log_file(self, filename_log: str, lsm_run_index, verbose=False):
        """Write log file of a selected LSM run.
        
        Parameters
        ----------
        filename_log : str
            Name and path of the output nsb file (e.g. /home/johnny/example.nsb)
        lsm_run_index : int
            Index of the lsm run in `campaign.lsm_runs` of which the results are exported to the nsb file.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        # Get and prepare data:
        lsm_run = self.lsm_runs[lsm_run_index]
        log_string = lsm_run.get_log_string

        # Append additional information to log file string:
        time_now_str = dt.datetime.now(tz=pytz.UTC).strftime('%Y-%m-%d, %H:%M:%S %Z')
        append_str = ''
        append_str += f'------------------------------------------\n'
        append_str += f'LSM run comment: {lsm_run.comment}\n'
        append_str += f'Log file created: {time_now_str}\n'
        append_str += f'Log file written with GravTools version: {GRAVTOOLS_VERSION}\n'
        out_string = log_string + '\n' + append_str
        
        # Write file:
        if verbose:
            print(f'Write log file to {filename_log}.')
        with open(filename_log, 'w') as out_file:
            out_file.write(out_string)

    def save_to_pickle(self, filename_pkl=None, verbose=True):
        """Save the campaign object to a pickle file at the given path.

        Parameters
        ----------
        filename_pkl : str, optional (default=`None`)
            Path and name of the pickle file, e.g. /home/user1/data/camp1.pkl. `None` indicates that the campaign object
            is saved to the default output directory past (`campaign.output_directory`). In this case the file is named
            `<campaign_name>.pkl`.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.

        Returns
        -------
        str : Name and path of the saved file.
        """
        if filename_pkl is None:
            filename_pkl = os.path.join(self.output_directory, f'{self.campaign_name}.pkl')
        # Open file:
        if verbose:
            print(f'Export campaign data to {filename_pkl}.')
        with open(filename_pkl, 'wb') as outfile:
            if PICKLE_PROTOCOL_VERSION == '999':
                pickle.dump(self, outfile, protocol=pickle.HIGHEST_PROTOCOL)
            else:
                pickle.dump(self, outfile, protocol=PICKLE_PROTOCOL_VERSION)
        return filename_pkl

    @classmethod
    def from_pkl(cls, filename_pkl: str, verbose: bool = True):
        """Loads a campaign object from a pickle file.

        Parameters
        ----------
        filename_pkl : str
            Name and path of the pickle file containing the campaign data.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.

        Returns
        -------
        `Campaign` object
        """
        campaign = pd.read_pickle(filename_pkl)
        if verbose:
            print(
                f'Loaded campaign "{campaign.campaign_name}" with {campaign.number_of_stations} station(s) and {campaign.number_of_surveys} survey(s).')
        campaign.check_survey_data(verbose=verbose)
        campaign.check_gravimeter_data(verbose=verbose)
        return campaign

    def check_gravimeter_data(self, verbose: bool = True):
        """Check gravimeter data in campaign and add missing information.

        Parameters
        ----------
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.

        Notes
        -----
        In case the gravimeters object is mission, it as added and initialized with data from the surveys.
        """
        if not hasattr(self, 'gravimeters'):
            self.init_gravimeters()
            if verbose:
                print('Missing gravimeter object added.')
            # Get gravimeter data from surveys:
            for survey_name, survey in self.surveys.items():
                self.gravimeters.add_from_survey(survey=survey, verbose=verbose)

    def check_survey_data(self, verbose: bool = True):
        """Check survey data in campaign, correct missing elements or raise an error.

        Parameters
        ----------
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        for survey_name, survey in self.surveys.items():
            # Check observation dataframe:
            is_valid, error_msg = survey.check_obs_df(verbose=verbose)
            if not is_valid:
                if verbose:
                    print('Error: ' + error_msg)
                raise RuntimeError(error_msg)
            # Check survey object attributes:
            survey.init_missing_attributes(verbose=verbose)

    def set_output_directory(self, output_directory):
        """Change the campaign's output directory.

        Parameters
        ----------
        output_directory : str
            New output directory. The specified directory has to exist on the computer/os!
        """
        # Check output directory:
        if not isinstance(output_directory, str):
            raise TypeError('The argument "output_directory" needs to be a string.')
        else:
            if not os.path.isdir(output_directory):
                raise AssertionError(f'The directory "{output_directory}" does not exist!')
        self.output_directory = output_directory

    def write_obs_list_csv(self, filename_csv: str, export_type: str = 'all_obs', verbose: bool = False):
        """Export a list of all observations in the current campaign.

        Parameters
        ----------
        filename_csv : str
            Name and path of the output CSV file.
        export_type : str, optional (default = 'all_obs')
            Defines which observations are exported. There are three options: (1) all observations ('all_obs'), (2) only
            active observations ('active_only') or (3) only inactive observations ('inactive_only').
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        # Prepare dataframe with all observations of all surveys
        export_survey_df_list = []
        for survey_name, survey in self.surveys.items():
            tmp_obs_df = survey.obs_df.copy(deep=True)
            tmp_obs_df.insert(0, 'survey_name', survey_name)
            export_survey_df_list.append(tmp_obs_df)
        export_obs_df = pd.concat(export_survey_df_list, ignore_index=True, sort=False)
        export_obs_df.sort_values('obs_epoch', inplace=True)

        # Filter data:
        if export_type != 'all_obs':
            if export_type == 'active_only':
                tmp_filter = export_obs_df['keep_obs']
            elif export_type == 'inactive_only':
                tmp_filter = ~export_obs_df['keep_obs']
            export_obs_df = export_obs_df.loc[tmp_filter, :]

        # Export to CSV file:
        if verbose:
            print(f'Write observation list to: {filename_csv}')
        export_obs_df.to_csv(filename_csv, index=False, columns=EXPORT_OBS_LIST_COLUMNS)

    def write_obs_list_of_lsm_run_csv(self, filename_csv: str, lsm_run_index: int,
                                      export_type: str = 'all_obs', verbose: bool = False):
        """Export a list of all observations that were used for calculating the setup data of a lsm_run.

        Parameters
        ----------
        filename_csv : str
            Name and path of the output CSV file.
        lsm_run_index : int
            Index of the lsm_run for which the observation list should be exported. If `None`, the current observation
            selection (in `Survey.obs_df`) is exported.
        export_type : str, optional (default = 'all_obs')
            Defines which observations are exported. There are three options: (1) all observations ('all_obs'), (2) only
            active observations ('active_only') or (3) only inactive observations ('inactive_only').
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.
        """
        # Get LSM run
        try:
            lsm_run = self.lsm_runs[lsm_run_index]
        except KeyError:
            raise RuntimeError(f'Invalid LSM run index: {lsm_run_index}')

        # Prepare dataframe with all observations of all surveys
        export_survey_df_list = []
        for survey_name, setup in lsm_run.setups.items():
            setup_obs_list_df = setup['setup_obs_list_df'].copy(deep=True)
            setup_obs_list_df.insert(0, 'survey_name', survey_name)
            export_survey_df_list.append(setup_obs_list_df)
        export_df = pd.concat(export_survey_df_list, ignore_index=True, sort=False)
        export_df.sort_values('obs_epoch', inplace=True)

        # Merge with survey data, if available:
        obs_df_list = []
        for survey_name, survey in self.surveys.items():
            obs_df = self.surveys[survey_name].obs_df.copy(deep=True)
            obs_df_list.append(obs_df)
        obs_df_all = pd.concat(obs_df_list, ignore_index=True, sort=False)
        obs_df_all.sort_values('obs_epoch', inplace=True)

        export_df = pd.merge(export_df, obs_df_all, how='left', left_on=['station_name', 'obs_epoch'],
                             right_on=['station_name', 'obs_epoch']).rename(
            columns={'keep_obs_x': 'keep_obs', 'station_name_x': 'station_name', 'obs_epoch_x': 'obs_epoch'})

        drop_col_list = list(set(export_df.columns) - set(EXPORT_OBS_LIST_COLUMNS))
        export_df.drop(columns=drop_col_list, inplace=True)

        # Filter data:
        if export_type != 'all_obs':
            if export_type == 'active_only':
                tmp_filter = export_df['keep_obs']
            elif export_type == 'inactive_only':
                tmp_filter = ~export_df['keep_obs']
            export_df = export_df.loc[tmp_filter, :]

        # Export to CSV file:
        if verbose:
            print(f'Write observation list to: {filename_csv}')
        export_df.to_csv(filename_csv, index=False, columns=EXPORT_OBS_LIST_COLUMNS)

    def flag_observations_based_on_obs_list_csv_file(self, obs_list_filename: str, update_type: str = 'all_obs',
                                                     verbose: bool = False):
        """ Flag observations in campaign based on an observation list file.

        Parameters
        ----------
        obs_list_filename : str
            Name and path of the input CSV file containing the observation list.
        update_type : str, optional (default = 'all')
            Defines which observations in the input list are used to update the `keep_obs` status of the matched
            observations in the campaign:. There are 3 options: (1) `all _obs` indicates that all matched observations
            are updated, (2) `inactive_only` indicates that only inactive observations (in the list) are updated and (3)
            `active_only` indicates that only active observations in the list are updated.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line.

        """
        if verbose:
            print(f'Flag observations based on observation list in: {obs_list_filename}')
        flag_log_str = ''

        # Read csv file:
        obs_list_df = pd.read_csv(obs_list_filename)
        if len(obs_list_df) > 0:
            # Check availability of needed columns:
            # EXPORT_OBS_LIST_COLUMNS
            invalid_cols = list(set(obs_list_df.columns) - set(EXPORT_OBS_LIST_COLUMNS))
            if len(invalid_cols) > 0:
                raise AssertionError(f'Invalid columns in the observation list csv file: {", ".join(invalid_cols)}')

            # Get filter for observations to be updated according to the "update_type":
            if update_type == 'all_obs':
                pass
            elif update_type == 'inactive_only':
                obs_list_df = obs_list_df[~obs_list_df['keep_obs']]
            elif update_type == 'active_only':
                obs_list_df = obs_list_df[obs_list_df['keep_obs']]
            else:
                raise AssertionError(f'Invalid input argument for "update_type": {update_type}')

            # Apply flagging:
            surveys = obs_list_df['survey_name'].unique().tolist()
            for survey_name in surveys:
                count_changed = 0
                count_matched = 0
                flag_log_str += f'Survey: {survey_name}\n'
                if survey_name not in self.surveys:
                    flag_log_str += '  - Does not exist in this campaign.\n'
                else:
                    num_obs_in_survey = len(obs_list_df[obs_list_df['survey_name'] == survey_name])
                    for index, row in obs_list_df[obs_list_df['survey_name'] == survey_name].iterrows():
                        obs_epoch = row['obs_epoch']
                        if sys.version_info >= (3, 7):
                            epoch_dt = dt.datetime.fromisoformat(obs_epoch)
                        else:
                            if (len(obs_epoch) == 25) & (obs_epoch[-3] == ':'):  # E.g.: '2018-10-23 05:52:01+00:00'
                                epoch_dt = dt.datetime.strptime(obs_epoch[:-3] + obs_epoch[-2:], '%Y-%m-%d %H:%M:%S%z')
                            else:
                                raise AssertionError('Unknown tim format!')
                        filter_tmp = (self.surveys[survey_name].obs_df['obs_epoch'] == epoch_dt) & (
                                    self.surveys[survey_name].obs_df['station_name'] == row['station_name'])
                        num_matched_rows = len(filter_tmp[filter_tmp])
                        if num_matched_rows > 1:
                            raise AssertionError(f'In survey {survey_name} the are {num_matched_rows} observations at the '
                                                 f'same time and station!')
                        if num_matched_rows == 1:  # OK!
                            count_matched += 1
                            if self.surveys[survey_name].obs_df.loc[filter_tmp, 'keep_obs'].bool() != row['keep_obs']:
                                self.surveys[survey_name].obs_df.loc[filter_tmp, 'keep_obs'] = row['keep_obs']
                                count_changed += 1
                    flag_log_str += f'  - Matched observations: {count_matched} of {num_obs_in_survey}\n'
                    flag_log_str += f'  - Changed "keep_obs" flags: {count_changed}\n'
                    # Activate/deactivate survey according to the keep_obs flag of their observations:

        else:
            flag_log_str = 'Empty observation list file!'
        if verbose:
            print(flag_log_str)
        return flag_log_str

    def change_campaign_name(self, name: str):
        """Change the name of the campaign.

        Parameters
        ----------
        name : str
            New campaign name. Non-empty string.
        """
        if isinstance(name, str):
            if name:  # Non-empty string
                if ' ' in name:
                    raise AssertionError(f'Blanks in the campaign name are not allowed!')
                else:
                    self.campaign_name = name
            else:
                raise AssertionError('The campaign name is emtpy!')
        else:
            raise AssertionError('The campaign name is not a string!')

    def __str__(self):
        return f'Campaign "{self.campaign_name}" with {self.number_of_surveys} surveys ' \
               f'and {self.stations.get_number_of_stations} stations.'


if __name__ == '__main__':
    """Main function, primarily for debugging and testing."""
    filename = '/pyProjects/gravtools/out/test.pkl'
    camp = Campaign.from_pkl(filename)



"""GravTools settings defined by the user.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Valid gravimeter data source types:
GRAVIMETER_DATA_SOURCE_TYPES = {
    'file': 'Gravimeter data loaded from file.',
    'survey': 'Default initialization because gravimeter was used in a survey without providing additional information',
}

# Default linear gravimeter scale factor. This value is used if no other information is provided.
DEFAULT_GRAVIMETER_LINEAR_SCALE_FACTOR = 1.0

# Additive constant for the determination of the full absolute gravity [µGal] from observed values:
ADDITIVE_CONST_ABS_GRAVITY = 9.8e8

SURVEY_DATA_SOURCE_TYPES = {
    'cg5_obs_file_txt': 'Scintrex CG5 observation file (text format)',
    'bev_obs_file': 'Simple observation file format used by BEV',
}
CG6_SURVEY_DATA_SOURCE_TYPES = {
    'cg6_obs_file_lynx_v1': 'Scintrex CG6 observation file (from Lynx LG, version 1, no notes)',
    'cg6_obs_file_lynx_v2': 'Scintrex CG6 observation file (from Lynx LG, version 2, with notes)',
    'cg6_obs_file_solo': 'Scintrex CG6 observation file (filtered data file from instrument)',
}
CG6_SURVEY_DATA_SOURCE_TYPES_SHORT = {
    'cg6_obs_file_lynx_v1': 'LynxLG, v1, no notes',
    'cg6_obs_file_lynx_v2': 'LynxLG, v2, with notes',
    'cg6_obs_file_solo': 'CG6 solo',
}
SURVEY_DATA_SOURCE_TYPES = {**SURVEY_DATA_SOURCE_TYPES, **CG6_SURVEY_DATA_SOURCE_TYPES}
CG6_SURVEY_DATA_SOURCE_TYPE_DEFAULT = 'cg6_obs_file_lynx_v2'

# Options, in which data column of the observation file pressure may be provided::
CG6_PRESSURE_IN_COLUMN = {
    'line': {'tooltip': 'Use the Line column to provide in-situ pressure data.',
             'label': 'Line',
             'formats': ['cg6_obs_file_lynx_v1', 'cg6_obs_file_lynx_v2', 'cg6_obs_file_solo']},
    'no_data': {'tooltip': 'No in-situ pressure data provided.',
                'label': ' ',
                'formats': ['cg6_obs_file_lynx_v1', 'cg6_obs_file_lynx_v2', 'cg6_obs_file_solo']},
}
CG6_PRESSURE_IN_COLUMN_DEFAULT = 'no_data'

# Options, in which data column of the observation file height differences to the ground (dhb) may be provided:
# items tooltips:
CG6_DHB_IN_COLUMN = {
    'occupation': {'tooltip': 'Use the Occupation column to provide height differences dhb.',
                   'label': 'Occupation (LynxLG only)',
                   'formats': ['cg6_obs_file_lynx_v1', 'cg6_obs_file_lynx_v2']},
    'line': {'tooltip': 'Use the Line column to provide height differences dhb.',
             'label': 'Line',
             'formats': ['cg6_obs_file_lynx_v1', 'cg6_obs_file_lynx_v2', 'cg6_obs_file_solo']},
    'no_data': {'tooltip': 'No height differences between the CG6 and the ground provided.',
                'label': ' ',
                'formats': ['cg6_obs_file_lynx_v1', 'cg6_obs_file_lynx_v2', 'cg6_obs_file_solo']},
}
CG6_DHB_IN_COLUMN_DEFAULT = 'no_data'

# Available options for location and error information provided in the CG6 observation files:
CG6_OBS_FILE_LOCATION_TYPES = {
    'gps': {'label': 'GPS',
            'tooltip': 'Coordinates determined with the build-in GPS module.'},
    'user': {'label': 'User',
             'tooltip': 'User defined coordinates.'},
}
CG6_OBS_FILE_LOCATION_TYPES_DEFAULT = 'user'
CG6_OBS_FILE_ERROR_TYPES = {
    'sd': {'label': 'Standard deviation',
           'tooltip': 'Standard deviation'},
    'se': {'label': 'Standard error',
           'tooltip': 'Standard error'},
}
CG6_OBS_FILE_ERROR_TYPES_DEFAULT = 'sd'

STATION_DATA_SOURCE_TYPES = {
    'oesgn_table': 'Austrian gravity base network stations (OESGN).',
    'csv_stat_file': 'CVS station file',
    'obs_file': 'From an observation file',
}

TIDE_CORRECTION_TYPES = {
    'instrumental_corr': 'Instrument-implemented tidal correction',
    'longman1959': 'Tidal corrections by the model of Longman (1959)',
    'no_tide_corr': 'No tide correction applied',
    'from_time_series': 'Interpolated from correction time series',
    'unknown': 'Unknown whether a tide correction was applied',
}

OCEANLOAD_CORRECTION_TYPES = {
    'no_oceanload_corr': 'No ocean-loading correction applied',
    'instrumental_corr': 'Instrument-implemented ocean-loading correction',
}

ATM_PRES_CORRECTION_TYPES = {
    'no_atm_pres_corr': 'No atmospheric pressure correction applied',
    'iso_2533_1975': 'Atmospheric pressure correction using ISO 2533:1975 normal air pressure',
}

SCALE_CORRECTION_TYPES = {
    'no_scale': 'No scaling applied',
    'linear_scale': 'Linear scaling applied',
}

# Default admittance factor for the calculation of atmospheric pressure correction using the ISO 2533:1975 normal air
# pressure
ATM_PRES_CORRECTION_ADMITTANCE_DEFAULT = 0.3

# Conversion factors of different units of gravity to µGal:
# - Keys: Units as used e.g. in TSF files by Tsoft
# - Values: Multiplicative conversion factors to µGal
UNIT_CONVERSION_TO_MUGAL = {
    'nm/s^2': 1e-1  # 1 nm/s^2 * 0.1 = 1 µGal
}

# Interpolation methods provided by scipy.interpolation.interp1, e.g. used for the interpolation of correction time
# series data.
SCIPY_INTERP1_INTERPOLATION_METHODS = {
    'linear': 'Linear interpolation.',
    'nearest': '"Snaps" to the nearest data point. Rounds down at half-integers (e.g. 0.5, 1.5)',
    'nearest-up': '"Snaps" to the nearest data point. Rounds up at half-integers (e.g. 0.5, 1.5)',
    'zero': 'Zero order spline. It`s value at any point is the last raw value seen.',
    'slinear': 'First order spline interpolation (similar to "linear").',
    'quadratic': 'Second order spline interpolation.',
    'cubic': 'Third order spline interpolation.',
    'previous': 'Return previous value.',
    'next': 'Return next value.',
}
# Default interpolation method:
SCIPY_INTERP1_INTERPOLATION_DEFAULT_METHOD = 'quadratic'

# Drift calculation methods (dialog observations/drift )
DRIFT_CALC_METHODS = {
    'numpy.polyfit': 'Least squares polynomial fit with numpy.polyfit().'
}

DRIFT_CALC_DEFAULT_METHOD = 'numpy_polyfit'

REFERENCE_HEIGHT_TYPE = {
    'sensor_height': 'The gravity values refer to the sensor height',
    'instrument_top': 'The gravity values refers to the instrument top',
    'ground': 'The gravity values refers to the ground point of the station',
    'control_point': 'The gravity values refers to the control point of the station',
}

# ##### Gravimeter default data #####
# The default information is used if no data is loaded from a gravimeter files (json)

# Default height offset between instrument reference surface (top) and the sensor level for different gravimeter types:
# - These default values are only used if no values are loaded from a gravimeter file
DEFAULT_GRAVIMETER_REFERENCE_HEIGHT_OFFSET_M = {
    'CG3': -0.211,
    'CG5': -0.211,
    'CG6': -0.099,  # W.r.t. top surface at the front
}

# Gravimeter default type
DEFAULT_GRAVIMETER_TYPE = 'n/a'

# Gravimeter default type
DEFAULT_GRAVIMETER_SERIAL_NUMBER = 'n/a'

# Default Gravimeter Type when loading data from an CG5 observation file
DEFAULT_GRAVIMETER_TYPE_CG5_SURVEY = 'CG5'

# Default Gravimeter Type when loading data from an CG6 observation file
DEFAULT_GRAVIMETER_TYPE_CG6_SURVEY = 'CG6'

# Default gravimeter description string:
DEFAULT_GRAVIMETER_DESCRIPTION = 'Default gravimeter settings'

# Default gravimeter calibration data:
# - Used if no data is provided through gravimeter files (json)
DEFAULT_CALIBRATION_START_DATE = '1900-01-01'
DEFAULT_CALIBRATION_END_DATE = '2100-01-01'
DEFAULT_CALIBRATION_LINEAR_FACTOR = 1.0
DEFAULT_CALIBRATION_COMMENT = 'default'

# Lookup table to convert gravimeter type to one-letter codes:
# - The one-letter codes are used in the nsb files (results). Nowhere else.
DEFAULT_GRAVIMETER_ONE_LETTER_CODES = {
    'CG5': 'C',
    'CG3': 'C',
    'CG6': 'C',
    DEFAULT_GRAVIMETER_TYPE: '?',
}

# Default manufacturers of gravimeter types
DEFAULT_GRAVIMETER_MANUFACTURERS = {
    'CG5': 'Scintrex',
    'CG3': 'Scintrex',
    'CG6': 'Scintrex',
    DEFAULT_GRAVIMETER_TYPE: 'unknown',
}

# Supported gravimeter types and description.
# - All supported gravimeters have to be listed here!
# - Gravimeter types stored in survey objects have to be listed here.
GRAVIMETER_TYPES = {
    'CG5': 'Sctintrex CG5',
    'CG3': 'Sctintrex CG3',
    'CG6': 'Sctintrex CG6',
    DEFAULT_GRAVIMETER_TYPE: 'Unknown gravimeter type',
}

#  Lookup table for matching gravimeter IDs and the tidal corrections that are applied per default in the BEV legacy
#  observation files:
# - Only required when loading "BEV legacy observation files", because there is no info on the applied tidal correction
BEV_GRAVIMETER_TIDE_CORR_LOOKUP = {
    '5': 'instrumental_corr'
}

# Methods for calculation of setup observations:
SETUP_CALC_METHODS = {
    'variance_weighted_mean': 'Variance weighted mean of observations within a setup.',
    'individual_obs': 'Each observation is treated as setup.'
}

# Methods for calculation of standard deviations (SD) of setup observations:
SETUP_SD_METHODS = {
    'sd_from_obs_file': 'Apply the standard deviations from observation files.',
    'sd_default_per_obs': 'Apply the default standard deviation to all individual observations.',
    'sd_default_per_setup': 'Apply the default standard deviation to setup observations.'
}

# Available adjustment methods:
ADJUSTMENT_METHODS = {
    'LSM_non_diff': 'LSM (non-differential observations)',  # Least-squares adjustment of non-differential observations
    'LSM_diff': 'LSM (differential observations)',  # Least-squares adjustment of differential observations
    'MLR_BEV': 'MLR (BEV legacy processing)',  # Least-squares adjustment of differential observations
    'VG_LSM_nondiff': 'VG LSM (non-differential observations)',
    # Vertical Gravity Gradient estimation based on Least-squares adjustment of non-differential observations
}

# List of LSM methods where export of nsb files is allowed:
LSM_METHODS_NSD_FILE_EXPORT = [
    'LSM_diff',
    'LSM_non_diff',
    'MLR_BEV',
]

# List of LSM methods where export of gis files is allowed:
LSM_METHODS_GIS_EXPORT = [
    'LSM_diff',
    'LSM_non_diff',
]

# Default subdirectory in the campaign's output directory for exporting the results as shapefiles:
GIS_RESULTS_OUTPUT_SUBDIR = 'GIS_results'

# List of LSM methods where VG plots are created (and should be saved as PNG files):
LSM_METHODS_VG_PLOT = [
    'VG_LSM_nondiff',
]

# Available iteration approaches for scaling the SD of setup observations:
ITERATION_APPROACHES = {
    'Multiplicative': 'Multiplicative iteration approach',
    'Additive': 'Additive iteration approach'
}

# Absolute tolerance for testing whether a value is equal to zero:
IS_ZERO_ABS_TOLERANCE = 1e-6

# Threshold for the "Gewichtsreziprokenprobe nach Ansermet" (see Skriptum AG1, p. 136, Eq. (6.86))
ANSERMET_DIFF_THRESHOLD = 1e-3

# Threshold for the redundancy component of an observation in order to apply a pope test for outlier detection:
R_POPE_TEST_THRESHOLD = 1e-6

# GUI and Program options:
CALCULATE_REDUCED_OBS_WHEN_LOADING_DATA = True  # Calculate reduced observations when loading observation data.
INIT_OESGN_STATION_AS_DATUM = False  # Initialize OESGN stations as datum stations when loading rom an OESGN file?

# ----- GUI appearance and plot settings -----

# Methods for the determination of histogram bin edges:
NUMPY_HISTOGRAM_BIN_EDGES_OPTIONS = {
    'auto': 'Maximum of the ‘sturges’ and ‘fd’ estimators. Provides good all around performance.',
    'fd': 'Freedman Diaconis Estimator: Robust (resilient to outliers) estimator that takes into account data variability and data size.',
    'doane': 'An improved version of Sturges’ estimator that works better with non-normal datasets.',
    'scott': 'Estimator based on leave-one-out cross-validation estimate of the integrated squared error. Can be regarded as a generalization of Scott’s rule.',
    'rice': 'Estimator does not take variability into account, only data size. Commonly overestimates number of bins required.',
    'sturges': 'R’s default method, only accounts for data size. Only optimal for gaussian data and underestimates number of bins for large non-gaussian datasets.',
    'sqrt': 'Square root (of data size) estimator, used by Excel and other programs for its speed and simplicity.',
    'Num. of bins': 'User defined number of bins (min. = 2, max. = 1000).'
}

# Maximum number of bina allowed when using "auto" or "fd" binning method:
HIST_MAX_BIN_NUM = 100

# Backup binning method if more than HIST_MAX_BIN_NUM bins are created:
HIST_BACKUP_BIN_METHOD = 'doane'

# Time label format for y-ticks in time-series plots:
# - Format string for datetime.strftime()
Y_TICK_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'

# --- General color settings ---
DATUM_STATION_COLOR = (255, 204, 204)

# --- Drift plots in the results tab: ---
# Number of plot items for plotting the drift function (polynomial):
DRIFT_PLOT_NUM_ITEMS_IN_DRIFT_FUNCTION = 100
# Options for the observation data points in the drift plot:
DRIFT_PLOT_SCATTER_PLOT_SYMBOL_SIZE = 10
DRIFT_PLOT_SCATTER_PLOT_PEN_WIDTH = 1
DRIFT_PLOT_SCATTER_PLOT_PEN_COLOR = 'k'

# --- Correlation matrix in the results tab: ---
# Background color scheme for the correlation coefficient [0 - 1]
# - Get HEX colors from here: https://colorbrewer2.org/
# - https://www.pythonguis.com/tutorials/qtableview-modelviews-numpy-pandas/
# green (low correlation) => red (high correlation)
CORRELATION_COEF_COLORS = ['#006837', '#1a9850', '#66bd63', '#a6d96a', '#d9ef8b', '#fee08b', '#fdae61', '#f46d43',
                           '#d73027', '#a50026']
# Background color for diagonal elements (=1):
CORRELATION_COEF_DIAG_ELEMENTS = '#bababa'  # light grey

# --- VG plot in the results tab: ---
# Min. and max. height for the horizontal plot axis [m]:
# - "None" indicates that the min/max of the horizontal axis is adjusted to the actual setup height (dhf) of the
# observations. In this case the plotted height range is widened by "VG_PLOT_HEIGHT_DELTA_M".
VG_PLOT_MIN_HEIGHT_M = 0.0
VG_PLOT_MAX_HEIGHT_M = 1.8
VG_PLOT_HEIGHT_DELTA_M = 0.1
# Number of datapoints plotted in the range between min. and max. height:
VG_PLOT_NUM_ITEMS_VG_POLYNOMIAL = 100
# Min. Y-range (default: [VG_PLOT_MIN_LOWER_L_RANGE, VG_PLOT_MIN_UPPER_Y_RANGE] = [-1, +1]):
VG_PLOT_MIN_UPPER_Y_RANGE = 1.0
VG_PLOT_MIN_LOWER_L_RANGE = -1.0
# Define pen for scatter plot of residuals:
VG_PLOT_SCATTER_PLOT_PEN_COLOR = 'k'
VG_PLOT_SCATTER_PLOT_PEN_WIDTH = 1
VG_PLOT_SCATTER_PLOT_SYMBOL_SIZE = 10

# ----- Data export options: -----
# List of columns in the `obs_df` dataframe that are written to the exported observation list CSV file:
EXPORT_OBS_LIST_COLUMNS = ['survey_name', 'station_name', 'obs_epoch', 'keep_obs']
# Maximum allowed SD of the estimated gravity at stations when exporting data to a nsb file!
#  => This is important as only 3 characters are reserved in the nsb file for the SD!
MAX_SD_FOR_EXPORT_TO_NSB_FILE = 999.0

# Choices for 5-character comment written to the nsb file (appears as Operat-G in the NSDB):
# - Official 5-character serial number of the Scintrex CG-5 Gravimeters: 'cg5_serial_number'
# - Version of Gravtools: 'gravtools_version'
# !! Warning: Actually only the gravtools version makes sense, because multiple surveys observed with different
# instruments can be combined
WRITE_COMMENT_TO_NSB = 'gravtools_version'

# ----- Program and Software settings -----
# Specify the protocol version of pickle used when saving the campaign data to a pickle file.
# Python 3.6.8 (used at BEV) does not support version 5, only up to version 4
# Options (type: int):
# - <int number> => Directly specify the protocol version
#                   (4 should be a good choice, as it is compatible with python 3.6)
# - 999 => Use the highest version available at the installation (pickle.HIGHEST_VERSION)
PICKLE_PROTOCOL_VERSION = 4

# ----- GIS data export settings: -----
# DEFAULT_EPSG_CODE = 4312  # 4312: MGI, lat/lon, Greenwich
DEFAULT_EPSG_CODE = 4326  # 4326: WGS84h
# Default filenames for shapefile export from the results tab:
DEFAULT_FILENAME_OBSERVATION_RESULTS_SHP = 'obs_results_'  # + <LSM run method>.shp
DEFAULT_FILENAME_STATION_RESULTS_SHP = 'stat_results_'  # + <LSM run method>.shp

# ----- SCHWAUS and DRIFT settings (legacy code) -----

# Instrumenten-IDs in der Messdatei einem Instrument zuweisen:
GRAVIMETER_ID_BEV = {
    '3': 'CG3',
    '5': 'CG5',
}

# Instrumenten-IDs in der Messdatei einem Instrumenten-KZ zuweisen:
GRAVIMETER_KZ_BEV = {
    '3': 'C',
    '5': 'C',
    '9': 'D',
    '90': 'D',
    '900': 'D',
    '51': 'D',
    '510': 'D',
    '500': 'W',
}

FLAG_SAVE_DRIFT_PLOT_PDF = True
FLAG_CREATE_SCHWAUS_PROTOCOL = True
VERBOSE = True

# Output directory
OUT_PATH = '/home/heller/pyProjects/GravTools/out'

# Default names and paths of input files:

# ÖSGN Table:
PATH_OESGN_TABLE = '/home/heller/pyProjects/GravTools/data/'
NAME_OESGN_TABLE = 'OESGN.TAB'

# BEV observation files:
PATH_OBS_FILE_BEV = '/home/heller/pyProjects/GravTools/data/BEV/'
NAME_OBS_FILE_BEV = 'f200701_1'

# CG-5 observation files (text)
PATH_OBS_FILE_CG5 = '/home/heller/pyProjects/GravTools/data/CG5/'
NAME_OBS_FILE_CG5 = '20200907_test.TXT'

# ##### Debug options #####
DEBUG_TIME_IT = False
